"""
This module is used to display the Density of States (DoS).
"""

from futile.Utils import write as safe_print

AU_eV = 27.21138386


class DiracSuperposition():
    """
    Defines as superposition of Dirac deltas which can be used to
    plot the density of states
    """

    def __init__(self, dos, wgts=[1.0]):
        """
        Parameters:
           dos: array containing the density of states per eack k-point.
                Should be of shape 2
           wgts: contains the weights of each of the k-points
        """
        import numpy as np
        self.dos = dos
        if isinstance(wgts, float):
            self.norm = [wgts]
        else:
            self.norm = wgts
        # set range for this distribution
        e_min = 1.e100
        e_max = -1.e100
        ddos = np.ravel(dos)
        if len(ddos) > 0:
            e_min = min(e_min, np.min(ddos) - 0.05 *
                        (np.max(ddos) - np.min(ddos)))
            e_max = max(e_max, np.max(ddos) + 0.05 *
                        (np.max(ddos) - np.min(ddos)))
        self.xlim = (e_min, e_max)

    def curve(self, xs, sigma, wgts=None):
        import numpy as np
        dos_g = 0.0
        idos = 0
        for norm, dos in zip(self.norm, self.dos):
            if wgts is not None:
                norms = wgts[idos]*norm
                idos += 1
            else:
                norms = np.ones(len(dos))*norm
            kptcurve = self.peaks(xs, dos, norms, sigma)
            dos_g += kptcurve
        return xs, dos_g

    def peak(self, omega, e, sigma):
        """
        Define if a peak is a Gaussian or a Lorenzian (temporarily only the
        gaussian is defined)
        """
        import numpy as np
        nfac = np.sqrt(2.0*np.pi)
        val = np.exp(- (omega - e)**2 / (2.0 * sigma**2))/(nfac*sigma)
        return val

    def peaks(self, xs, dos, norms, sigma):
        """
        Return the array of the whole set of peaks
        """
        curve = 0.0
        for e, nrm in zip(dos, norms):
            curve += self.peak(xs, e, sigma)*nrm
        return curve


def _bandarray_to_data(jspin, bandarrays):
    lbl = 'up' if jspin == 0 else 'dw'
    kptlists = [[], []]
    for orbs in bandarrays:
        for ispin, norbs in enumerate(orbs.info):
            if norbs == 0 or ispin != jspin:
                continue
            # energy values
            kptlists[0].append(orbs[ispin, :norbs])
            # normalization
            kptlists[1].append(orbs.kwgt*(1.0-2*ispin))
            # print 'kpt',kptlists
    return kptlists, lbl


class DoS():
    """
    Definition of the density of state class
    """

    def __init__(self, bandarrays=None, energies=None, evals=None, units='eV',
                 label='1', sigma=0.2/AU_eV, npts=2500, fermi_level=None,
                 norm=1.0, sdos=None):
        """
        Extract a quantity which is associated to the DoS, that can be plotted
        """
        import numpy as np
        self.ens = []
        self.labels = []
        # self.norms=[]
        self.npts = npts
        if bandarrays is not None:
            self.append_from_bandarray(bandarrays, label)
        if evals is not None:
            self.append_from_dict(evals, label)
        if energies is not None:
            self.append(np.array([energies]), label=label, units=units,
                        norm=(np.array([norm])
                        if isinstance(norm, float) else norm))
        self.sigma = self.conversion_factor(units)*sigma
        self.fermi_level(fermi_level, units=units)
        if sdos is not None:
            self._embed_sdos(sdos)

    def _embed_sdos(self, sdos):
        self.sdos = []
        for i, xdos in enumerate(sdos):
            self.sdos.append({'coord': xdos['coord']})
            jdos = 0
            for subspin in xdos['dos']:
                if len(subspin[0]) == 0:
                    continue
                d = {'doslist': subspin}
                try:
                    self.ens[jdos]['sdos'].append(d)
                except KeyError:
                    self.ens[jdos]['sdos'] = [d]
                jdos += 1

    def append_from_bandarray(self, bandarrays, label):
        """
        Add a new band array to the previous DoS. Important for kpoints DOS
        """
        import numpy as np
        for jspin in range(2):
            kptlists, lbl = _bandarray_to_data(jspin, bandarrays)
            self.append(np.array(kptlists[0]), label=label+lbl, units='AU',
                        norm=np.array(kptlists[1]))

    def append_from_dict(self, evals, label):
        import numpy as np
        "Get the energies from the different flavours given by the dict"
        evs = [[], []]
        ef = None
        for ev in evals:
            occ = self.get_ev(ev, ['e_occ', 'e_occupied'])
            if occ:
                ef = max(occ)
            vrt = self.get_ev(ev, ['e_vrt', 'e_virt'])
            eigen = False
            if occ:
                eigen = occ
            if vrt:
                eigen = vrt
            if not eigen:
                eigen = self.get_ev(ev)
            if not occ and not vrt and eigen:
                ef = max(eigen)
            if not eigen:
                continue
            for i, e in enumerate(eigen):
                if e:
                    evs[i].append(e)
        for i, energs in enumerate(evs):
            if len(energs) == 0:
                continue
            self.append(np.array(energs), label=label,
                        units='AU', norm=1.0-2.0*i)
        if ef:
            self.fermi_level(ef, units='AU')

    def get_ev(self, ev, keys=None):
        "Get the correct list of the energies for this eigenvalue"
        res = False
        if keys is None:
            ener = ev.get('e')
            spin = ev.get('s')
            if ener and spin == 1:
                res = [ener]
            elif ener and spin == -1:
                res = [None, ener]
        else:
            for k in keys:
                if k in ev:
                    res = ev[k]
                    if not isinstance(res, list):
                        res = [res]
                    break
        return res

    def append(self, energies, label=None, units='eV', norm=1.0):
        if not isinstance(norm, float) and len(norm) == 0:
            return
        dos = self.conversion_factor(units)*energies
        self.ens.append({'dos': DiracSuperposition(dos, wgts=norm)})
        lbl = label if label is not None else str(len(self.labels)+1)
        self.labels.append(lbl)
        # self.norms.append(norm)
        self.range = self._set_range()

    def conversion_factor(self, units):
        if units == 'AU':
            fac = AU_eV
        elif units == 'eV':
            fac = 1.0
        else:
            raise ValueError('Unrecognized units ('+units+')')
        return fac

    def fermi_level(self, fermi_level, units='eV'):
        if fermi_level is not None:
            self.ef = fermi_level*self.conversion_factor(units)

    def _set_range(self, npts=None):
        import numpy as np
        if npts is None:
            npts = self.npts
        e_min = 1.e100
        e_max = -1.e100
        for dos in self.ens:
            mn, mx = dos['dos'].xlim
            e_min = min(e_min, mn)
            e_max = max(e_max, mx)
        return np.arange(e_min, e_max, (e_max-e_min)/npts)

    def curve(self, dos, norm, sigma=None):
        import numpy as np
        if sigma is None:
            sigma = self.sigma
        nrm = np.sqrt(2.0*np.pi)*sigma/norm
        dos_g = []
        for e_i in self.range:
            if len(dos.shape) == 2:
                nkpt = dos.shape[0]
                value = 0.0
                for ikpt in range(nkpt):
                    value += np.sum(np.exp(- (e_i - dos[ikpt, :])**2 /
                                           (2.0 * sigma**2))/nrm[ikpt])
            else:
                value = np.sum(
                    np.exp(- (e_i - dos[:])**2 / (2.0 * sigma**2))/nrm)
            # Append data corresponding to each energy grid
            dos_g.append(value)
        return np.array(dos_g)

    def dump(self, sigma=None):
        "For Gnuplot"
        if sigma is None:
            sigma = self.sigma
        # data=[self.curve(dos,norm=self.norms[i],sigma=sigma)
        #       for i,dos in enumerate(self.ens)]
        data = [dos['dos'].curve(self.range, sigma=sigma)[1]
                for dos in self.ens]

        for i, e in enumerate(self.range):
            safe_print(e, ' '.join(map(str, [d[i] for d in data])))

    def plot(self, sigma=None, legend=True, xlmin=None, xlmax=None, ylmin=None,
             ylmax=None):
        import matplotlib.pyplot as plt
        from matplotlib.widgets import Slider  # , Button, RadioButtons
        if sigma is None:
            sigma = self.sigma
        self.fig, self.ax1 = plt.subplots()
        self.plotl = []
        for i, dos in enumerate(self.ens):
            # self.plotl.append(self.ax1.plot(self.range,self.curve(dos,norm=self.norms[i],sigma=sigma),label=self.labels[i]))
            self.plotl.append(self.ax1.plot(
                *dos['dos'].curve(self.range, sigma=sigma),
                label=self.labels[i]))
        if xlmax is not None:
            plt.xlim(xmax=xlmax)
        if xlmin is not None:
            plt.xlim(xmin=xlmin)
        if ylmax is not None:
            plt.ylim(ymax=ylmax)
        if ylmin is not None:
            plt.ylim(ymin=ylmin)
        plt.xlabel('Energy [eV]', fontsize=18)
        plt.ylabel('DoS', fontsize=18)
        if self.ef is not None:
            plt.axvline(self.ef, color='k', linestyle='--')
            # self.ax1.annotate('Fermi level', xy=(self.ef,2),
            #        xytext=(self.ef, 10),
            #    arrowprops=dict(facecolor='white', shrink=0.05),
            # )
        if len(self.labels) > 1 and legend:
            plt.legend(loc='best')
        axcolor = 'lightgoldenrodyellow'
        try:
            axsigma = plt.axes([0.2, 0.93, 0.65, 0.03], facecolor=axcolor)
        except AttributeError:
            axsigma = plt.axes([0.2, 0.93, 0.65, 0.03], axisbg=axcolor)

        self.ssig = Slider(axsigma, 'Smearing', 0.0, 0.4, valinit=sigma)
        self.ssig.on_changed(self.update)
        if hasattr(self, 'sdos') and self.sdos:
            self._set_sdos_selector()
            self._set_sdos()
        plt.show()

    def _set_sdos_selector(self):
        import matplotlib.pyplot as plt
        from matplotlib.widgets import RadioButtons
        self.sdos_selector = RadioButtons(
            plt.axes([0.93, 0.05, 0.04, 0.11], axisbg='lightgoldenrodyellow'),
            ('x', 'y', 'z'), active=1)
        self.isdos = 1
        self.sdos_selector.on_clicked(self._update_sdos)

    def _set_sdos(self):
        import numpy
        xs = self.sdos[self.isdos]['coord']
        self._set_sdos_sliders(numpy.min(xs), numpy.max(xs))
        self._update_sdos(0.0)  # fake value as it is unused

    def _sdos_curve(self, sdos, vmin, vmax):
        import numpy
        xs = self.sdos[self.isdos]['coord']
        imin = numpy.argmin(numpy.abs(xs-vmin))
        imax = numpy.argmin(numpy.abs(xs-vmax))
        doslist = sdos[self.isdos]['doslist']
        # norms=self.sdos[self.isdos]['norms'][ispin]
        tocurve = [0.0 for i in doslist[imin]]
        for d in doslist[imin:imax+1]:
            tocurve = [t+dd for t, dd in zip(tocurve, d)]
        # tocurve=numpy.sum([ d[ispin] for d in doslist[imin:imax+1]],axis=0)
        return tocurve
        # float(len(xs))/float(imax+1-imin)*tocurve,norms

    def _update_sdos(self, val):
        isdos = self.isdos
        if val == 'x':
            isdos = 0
        elif val == 'y':
            isdos = 1
        elif val == 'z':
            isdos = 2
        if isdos != self.isdos:
            self.isdos = isdos
            self._set_sdos()

        vmin, vmax = (s.val for s in self.ssdos)
        if vmax < vmin:
            self.ssdos[1].set_val(vmin)
            vmax = vmin
        if vmin > vmax:
            self.ssdos[0].set_val(vmax)
            vmin = vmax
        # now plot the sdos curve associated to the given value
        sig = self.ssig.val
        curves = []
        for dos in self.ens:
            if 'sdos' not in dos:
                continue
            renorms = self._sdos_curve(dos['sdos'], vmin, vmax)
            curve = dos['dos'].curve(self.range, sigma=sig, wgts=renorms)
            curves.append(curve)
        if hasattr(self, '_sdos_plots'):
            for pl, curve in zip(self._sdos_plots, curves):
                pl[0].set_ydata(curve[1])
        else:
            self._sdos_plots = []
            for c in curves:
                self._sdos_plots.append(
                    self.ax1.plot(*c, label='sdos'))
        self.ax1.relim()
        self.ax1.autoscale_view()
        self.fig.canvas.draw_idle()

    def _set_sdos_sliders(self, cmin, cmax):
        import matplotlib.pyplot as plt
        from futile.Figures import VertSlider
        if hasattr(self, 'ssdos'):
            self.ssdos[0].ax.clear()
            self.ssdos[0].__init__(
                self.ssdos[0].ax, 'SDos', cmin, cmax, valinit=cmin)
            self.ssdos[1].ax.clear()
            self.ssdos[1].__init__(self.ssdos[1].ax, '',
                                   cmin, cmax, valinit=cmax)
        else:
            axcolor = 'red'
            axmin = plt.axes([0.93, 0.2, 0.02, 0.65], axisbg=axcolor)
            axmax = plt.axes([0.95, 0.2, 0.02, 0.65], axisbg=axcolor)
            self.ssdos = [
                VertSlider(axmin, 'SDos', cmin, cmax, valinit=cmin),
                VertSlider(axmax, '', cmin, cmax, valinit=cmax)]
        self.ssdos[0].valtext.set_ha('right')
        self.ssdos[1].valtext.set_ha('left')
        self.ssdos[0].on_changed(self._update_sdos)
        self.ssdos[1].on_changed(self._update_sdos)

    def update(self, val):
        sig = self.ssig.val
        for i, dos in enumerate(self.ens):
            self.plotl[i][0].set_ydata(
                dos['dos'].curve(self.range, sigma=sig)[1])
            # self.plotl[i][0].set_ydata(self.curve(dos,norm=self.norms[i],sigma=sig))

        self.ax1.relim()
        self.ax1.autoscale_view()
        self.fig.canvas.draw_idle()


if __name__ == "__main__":
    import numpy as np
    energies = np.array([-0.815924953235059, -0.803163374736654,
                         -0.780540200987971, -0.7508806541364,
                         -0.723626807289917, -0.714924448617026,
                         -0.710448085701742, -0.68799028016451,
                         -0.67247569974853, -0.659038909236607,
                         -0.625396293324399, -0.608009041659988,
                         -0.565337910777367, -0.561250536074343,
                         -0.551767438323268, -0.541295070404525,
                         -0.532326667587434, -0.515961980147107,
                         -0.474601108285518, -0.473408476151224,
                         -0.46509070541069, -0.445709086452906,
                         -0.433874403837837, -0.416121660651406,
                         -0.407871082254237, -0.406123490618786,
                         -0.403004188319382, -0.38974739285104,
                         -0.380837488456638, -0.375163102271681,
                         -0.375007771592681, -0.367898783582561,
                         -0.367518948507212, -0.359401585874402,
                         -0.358189406008502, -0.354517727598174,
                         -0.334286389724978, -0.332921810616845,
                         -0.315466259109401, -0.308028853904577,
                         -0.29864142362141, -0.294024743731349,
                         -0.292104129933301, -0.285165738729842,
                         -0.28419932605141, -0.267399999874122,
                         -0.259487769142101, -0.239899780812716,
                         -0.224858003804207, -0.20448050758473,
                         -0.164155133452971, -0.117617164459898,
                         -0.0717938081884113, -0.0526986239898579,
                         -0.0346031190163735, -0.0167949342608791,
                         -0.0135168064347152, -0.0102971895842409,
                         0.00759271179427191, 0.00974950976249545,
                         0.010176021051287, 0.0217652761059223,
                         0.0239924727094222, 0.0413057846713024,
                         0.0422334333464529, 0.0459150454793617,
                         0.0517637894860314])
    dos = DoS(energies, fermi_level=-0.1)
    dos.append(0.2+energies)
    dos.dump(sigma=0.01)
    dos.plot()



# -*- coding: utf-8 -*-

"""
Calculations provided by aiida_bigdft.

Register calculations via the "aiida.calculations" entry point in setup.json.
"""
from __future__ import absolute_import

from aiida.common import datastructures
from aiida.engine import CalcJob
from aiida.orm.nodes.data import List
from aiida.plugins import DataFactory

RemoteData = DataFactory('remote')


class ScriptCalculation(CalcJob):
    """
    AiiDA calculation to add post treatments to a computation workcahin.
    post treatment scripts are to be registered as codes in aiida.
    They are python scripts accepting one argument : a remotefolder where data is stored
    Output files are not specified and can be added to the extra_retrieved_files list
    """

    @classmethod
    def define(cls, spec):
        """Define inputs and outputs of the calculation."""
        # yapf: disable
        super(ScriptCalculation, cls).define(spec)
        spec.input('metadata.options.resources', valid_type=dict, default={'num_machines': 1, 'num_mpiprocs_per_machine': 1})
        spec.input('bigdft_data_folder', valid_type=RemoteData, help='Folder to the BigDFT data folder')
        spec.input('retrieved_files', valid_type=List, help='', default=lambda: List())
        spec.exit_code(101, 'ERROR_SCRIPT',
                       'Script execution failed')


    def prepare_for_submission(self, folder):
        """
        Create input files.

        :param folder: an `aiida.common.folders.Folder` where the plugin should temporarily place all files needed by
            the calculation.
        :return: `aiida.common.datastructures.CalcInfo` instance
        """
        codeinfo = datastructures.CodeInfo()
        codeinfo.code_uuid = self.inputs.code.uuid
        if not self.inputs.code.can_run_on(
                self.inputs.bigdft_data_folder.computer):
            self.report("This post-processing script {}\
                         can't run on  {} where data resides",
                        format(self.inputs.code, self.inputs.bigdft_data_folder.get_computer_name()))
            return self.exit_codes.ERROR_SCRIPT

        codeinfo.withmpi = False
        codeinfo.cmdline_params = [self.inputs.bigdft_data_folder.get_remote_path()]
        # Prepare a `CalcInfo` to be returned to the engine
        calcinfo = datastructures.CalcInfo()

        calcinfo.codes_info = [codeinfo]
        calcinfo.local_copy_list = []
        calcinfo.retrieve_list = []
        calcinfo.retrieve_list.extend(self.inputs.retrieved_files)

        return calcinfo



""" Tests for calculations

"""
from __future__ import print_function
from __future__ import absolute_import

import os
from aiida_bigdft import tests


def test_process(aiida_code):
    """Test running a calculation
    note this does not test that the expected outputs are created of output parsing"""
    from aiida.plugins import DataFactory, CalculationFactory
    from aiida.engine import run
    #TODO



"""
This module contains some wrappers for using OpenBabel to perform
various operations on BigDFT molecules.

https://open-babel.readthedocs.io/en/latest/UseTheLibrary/Python.html
"""


def convert_atom_to_babel(atom):
    """
    Given a BigDFT atom, this routine creates the open babel
    equivalent.

    Args:
      atom (BigDFT.Atom.Atom): the atom to convert.

    Returns:
      (openbabel.OBAtom): an open babel atom type.
    """
    from openbabel import OBAtom
    a = OBAtom()

    pos = atom.get_position(units="angstroem")
    a.SetVector(pos[0], pos[1], pos[2])
    a.SetAtomicNum(atom.atomic_number)

    if atom.q0 is not None:
        a.SetPartialCharge(atom.q0)

    return a


def convert_frag_to_babel(frag, generate_connectivity=False):
    """
    Convert a BigDFT fragment to an open babel molecule.

    Args:
      frag (BigDFT.Fragments.Fragment): the fragment to convert.
      generate_connectivity (bool): if true, the connectivity will
        be generated by openbabel.

    Returns:
      (openbabel.OBMol): an open babel type molecule.
    """
    from openbabel import OBMol

    # Insert the atoms.
    mol = OBMol()
    for at in frag:
        obat = convert_atom_to_babel(at)
        mol.InsertAtom(obat)

    # Add the bonds.
    if frag.conmat is not None:
        for i in frag.conmat:
            for j in frag.conmat[i]:
                if i > j:
                    mol.AddBond(i+1, j+1, int(frag.conmat[i][j]))
        mol.PerceiveBondOrders()

    return mol


def convert_babel_to_frag(mol):
    """
    Convert an openbabel molecule to a BigDFT fragment.

    Args:
      mol (OBMol): the molecule to compute.

    Returns:
      (BigDFT.Fragments.Fragment): a BigDFT representation of the fragment.
    """
    from BigDFT import Atom as AA
    from BigDFT.Fragments import Fragment
    from openbabel import OBMolAtomIter

    frag = Fragment()
    for obat in OBMolAtomIter(mol):
        sym = AA.number_to_symbol(obat.AtomicNumber())
        pos = [obat.GetX(), obat.GetY(), obat.GetZ()]

        frag += [AA.Atom({sym: pos, "units": "angstroem"})]

    return frag


def compute_fragment_smiles(frag):
    """
    Computes the SMILES representation of a given fragment.

    Args:
      frag (BigDFT.Fragments.Fragment): the fragment to compute the
        representation of.

    Return:
      (str): the smiles representation of this molecule.
    """
    from openbabel import OBConversion

    conv = OBConversion()
    mol = convert_frag_to_babel(frag)
    conv.SetOutFormat("SMI")

    retstr = conv.WriteString(mol)
    retstr = retstr.replace("\n", "")
    retstr = retstr.replace("\t", "")

    return retstr


def compute_fragment_fingerprint(frag, fps="fp2"):
    """
    Computes the fingerprint for a particular fragment.

    Args:
      frag (BigDFT.Fragments.Fragment): the fragment to compute the
        representation of.
      fps (str): the type of finger print to compute.

    Return:
      (openbabel.OBFingerprint): a fingerprint for this fragment.
    """
    from pybel import Molecule

    mol = convert_frag_to_babel(frag)
    pybelmol = Molecule(mol)

    return pybelmol.calcfp(fps)


def optimize_fragment(frag, forcefield="MMFF94", method="SteepestDescent",
                      steps=1000, econv=1e-6, constraints=None):
    """
    Optimize the geometry of a given fragment.

    Args:
      frag (BigDFT.Fragments.Fragment): the fragment to optimize.
      forcefield (str): the type of forcefield to use.
      constraints (list): for each atom, list whether it if frozen or not.

    Returns:
      (BigDFT.Fragments.Fragment): a new fragment with the optimized positions.
    """
    from openbabel import OBForceField, OBMolAtomIter
    from copy import deepcopy

    # Setup the calculation
    ff = OBForceField.FindForceField(forcefield)
    mol = convert_frag_to_babel(frag, generate_connectivity=True)

    ff.Setup(mol)

    if constraints:
        constr = ff.GetConstraints()
        for i in range(0, len(constraints)):
            if constraints[i] is None:
                continue
            elif constraints[i] == "fxyz" or constraints[i] == "f":
                constr.AddAtomConstraint(i+1)
            elif constraints[i] == "fx":
                constr.AddAtomXConstraint(i+1)
            elif constraints[i] == "fy":
                constr.AddAtomYConstraint(i+1)
            elif constraints[i] == "fz":
                constr.AddAtomZConstraint(i+1)
        ff.SetConstraints(constr)

    # Call the optimization routine.
    if method == "SteepestDescent":
        ff.SteepestDescent(steps, econv)
    elif method == "ConjugateGradients":
        ff.ConjugateGradients(steps, econv)
    else:
        raise ValueError("Invalid minimization method.")

    # Extract out the the positions.
    ff.GetCoordinates(mol)
    newfrag = deepcopy(frag)
    for at, obat in zip(newfrag, OBMolAtomIter(mol)):
        at.set_position([obat.GetX(), obat.GetY(), obat.GetZ()],
                        units="angstroem")

    return newfrag


def optimize_system(sys, forcefield="MMFF94", method="SteepestDescent",
                    steps=1000, econv=1e-6,):
    """
    Optimize the geometry of a given system.

    Args:
      sys (BigDFT.Fragments.System): the system to optimize.
      forcefield (str): the type of forcefield to use.

    Returns:
      (BigDFT.Fragments.System): a new system with the optimized positions.
    """
    from BigDFT.Fragments import Fragment
    from copy import deepcopy

    # Convert to one big system.
    fullfrag = Fragment()
    for frag in sys.values():
        for at in frag:
            fullfrag += [at]

    # Compute the connectivity, first the lookup table.
    if sys.conmat is not None:
        counter = 0
        lookup = {}
        for fragid, frag in sys.items():
            for i, at in enumerate(frag):
                lookup[(fragid, i)] = counter
                counter = counter + 1

        # Now the actual connectivity matrix.
        fullfrag.conmat = {}
        for fragid, frag in sys.items():
            for i in range(0, len(frag)):
                t1 = (fragid, i)
                t2list = sys.conmat[fragid][i]
                fullfrag.conmat[lookup[t1]] = {}
                for t2, bo in t2list.items():
                    fullfrag.conmat[lookup[t1]][lookup[t2]] = bo

    # Setup the constraints
    constraint_list = []
    counter = 0
    for frag in sys.values():
        for at in frag:
            constraint_list.append(frag.frozen)
            counter = counter + 1

    # Optimize
    optfrag = optimize_fragment(fullfrag, forcefield=forcefield, method=method,
                                steps=steps, econv=econv,
                                constraints=constraint_list)

    # Convert back
    outsys = deepcopy(sys)
    counter = 0
    for frag in outsys.values():
        for i in range(0, len(frag)):
            frag[i] = optfrag[counter]
            counter = counter + 1

    return outsys



"""
This module is related to the usage of BigDFT with Fragment-related Quantities.
Input as well as Logfiles might be processed with the classes and methods
provided by it.

The main two classes for this module are :class:`BigDFT.Fragments.System` and
:class:`BigDFT.Fragments.Fragment`. A System is a named collection of
fragments, and a Fragment is a list of atoms. Thus a System behaves much like
a dictionary, whereas a Fragment behaves more like a list. All of the basic
dictionary and list like operations can be applied.
"""

from futile.Utils import write as safe_print
try:
    from collections.abc import MutableMapping, MutableSequence
except ImportError:
    from collections import MutableMapping, MutableSequence


def GetFragTuple(fragid):
    """
    Fragment ids should have the form: "NAME:NUMBER". This splits the fragment
    into the name and number value.

    Args:
      fragid (str): the fragment id string.

    Return:
      (tuple): fragment name, fragment number
    """
    return (fragid.split(":")[0], fragid.split(":")[1])


def distance(i, j, cell=None):
    """
    Distance between fragments, defined as distance between center of mass

    Args:
      i (Fragment): first fragment.
      j (Fragment): second fragment.
      cell (array): an array describing the unit cell.

    Returns:
      (float): the distance between centers of mass.
    """
    import numpy
    vec = i.centroid - j.centroid
    if cell:
        per = numpy.where(cell > 0.0)
        for i, p in enumerate(per):
            vec -= cell[i] * int(round(vec[i] / cell[i]))
    return numpy.sqrt(numpy.dot(vec, vec.T))


def pairwise_distance(i, j, cell=None):
    """
    Distance between fragments, as defined by the distance between their two
    nearest atoms.

    Args:
      i (Fragment): first fragment.
      j (Fragment): second fragment.
      cell (array): an array describing the unit cell.

    Returns:
      (float): the pairwise distance between fragments.
    """
    if not cell:
        from scipy.spatial.distance import cdist
        pos_i = [at.get_position() for at in i]
        pos_j = [at.get_position() for at in j]

        dmat = cdist(pos_i, pos_j)

        dist = dmat.min()

    else:
        dist = distance(Fragment([i[0]]), Fragment([j[0]]), cell)
        for at_i in i:
            for at_j in j:
                new_d = distance(Fragment([at_i]), Fragment([at_j]), cell)
                if new_d < dist:
                    dist = new_d

    return dist


def plot_fragment_information(axs, datadict, colordict=None, minval=None):
    """
    Often times we want to plot measures related to the differnet fragments
    in a system. For this routine, you can pass a dictionary mappin
    fragment ids to some kind of value. This routine takes care of the
    formatting of the axis to easily read the different fragment names.

    Args:
      axs (matplotlib.Axes): an axes object to plot on.
      datadict (dict): a dictionary from fragment ids to some kind of data
        value.
      colordict (dict): optionally, a dictionary from fragment ids to a
        color value.
      minval (float): only values above this minimum value are plotted.
    """
    from BigDFT.Fragments import GetFragTuple

    # Compute minval
    if not minval:
        minval = min(datadict.values())

    # Sort by fragment id
    slabels = sorted(datadict.keys(),
                     key=lambda x: int(GetFragTuple(x)[1]))
    svalues = [datadict[x] for x in slabels]

    # Remove values below the minimum
    slabels = [x for x in slabels if datadict[x] >= minval]
    svalues = [x for x in svalues if x >= minval]

    # Label the axis by fragments
    axs.set_xlabel("Fragment", fontsize=12)
    axs.set_xticks(range(len(datadict.keys())))
    axs.set_xticklabels(slabels, rotation=90)

    # Plot the actual values
    axs.plot(svalues, 'x', markersize=12, color='k')

    # Plot the colored values.
    if colordict:
        for i, key in enumerate(slabels):
            if key not in colordict:
                continue
            axs.plot(i, svalues[i], 'x', markersize=12,
                     color=colordict[key])


class Lattice():
    """
    Defines the fundamental objects to deal with periodic systems
    """

    def __init__(self, vectors):
        self.vectors = vectors

    def grid(self, origin=[0.0, 0.0, 0.0], extremes=None, radius=None):
        """
        Produces a set of translation vectors from a given origin
        """
        import numpy as np
        transl = []
        g = [[], [], []]  # the grid of discrete translations
        if extremes is not None:
            # print extremes
            for i, d in enumerate(extremes):
                for k in range(d[0], d[1] + 1):
                    g[i].append(k)
            # print g
            for i in g[0]:
                arri = np.array(self.vectors[0]) * i
                for j in g[1]:
                    arrj = np.array(self.vectors[1]) * j + arri
                    for k in g[2]:
                        arrk = np.array(self.vectors[2]) * k + arrj
                        vect = np.array(origin) + arrk
                        app = True
                        if radius is not None:
                            app = np.linalg.norm(arrk) < radius
                        if app:
                            transl.append(vect)
        return transl


class RotoTranslation():
    """
    Define a transformation which can be applied to a fragment. This
    rotation is defined by giving this class two fragments, and the
    rototranslation between those fragments is automatically computed.

    Args:
      frag1 (BigDFT.Fragments.Fragment): the first position.
      frag2 (BigDFT.Fragments.Fragment): the second position.
    """
    def __init__(self, frag1, frag2):
        try:
            from BigDFT import wahba
            from numpy import matrix

            # Setup each fragment as a matrix of positions
            pos1 = matrix([at.get_position() for at in frag1])
            pos2 = matrix([at.get_position() for at in frag2])

            # Compute transformation
            self.R, self.t, self.J = wahba.rigid_transform_3D(pos1, pos2)
        except Exception as e:
            safe_print('Error', e)
            self.R, self.t, self.J = (None, None, 1.0e10)

    def dot(self, frag):
        """
        Apply the rototranslations on a fragment.

        Args:
          frag (BigDFT.Fragments.Fragment): the fragment to rototranslate.

        Return:
          (BigDFT.Fragments.Fragment): the rototranslated fragment.
        """
        from BigDFT import wahba as w
        from numpy import matrix, array
        from copy import deepcopy

        # Convert to a position matrix
        pos = matrix([at.get_position() for at in frag])

        # Apply roto-translation
        if self.t is None:
            res = w.apply_R(self.R, pos)
        elif self.R is None:
            res = w.apply_t(self.t, pos)
        else:
            res = w.apply_Rt(self.R, self.t, pos)

        # Copy back to the fragment
        newfrag = deepcopy(frag)
        for i in range(0, len(newfrag)):
            newfrag[i].set_position([float(x) for x in array(res[i, :])[0]])

        return newfrag

    def invert(self):
        """
        Computes the inverse rototranslation.
        """
        self.t = -self.t
        if self.R is not None:
            self.R = self.R.T


class Translation(RotoTranslation):
    """
    This class defines a simple translation.

    Args:
      t (list): the vector describing the translation.
    """
    def __init__(self, t):
        import numpy
        self.R = None
        self.t = numpy.mat(t).reshape(3, 1)
        self.J = 0.0


class Rotation(RotoTranslation):
    """
    This class defines a simple rotation.

    Args:
      t (list): the vector describing the rotation.
    """
    def __init__(self, R):
        self.t = None
        self.R = R
        self.J = 0.0


def interpolate_fragments(A, B, steps, extrapolation_steps=0):
    """
    Given two fragments A and B, this generates a list of Fragments
    that interpolate between A and B in a specified number of steps.

    Args:
      A (BigDFT.Fragments.Fragment): starting fragment.
      B (BigDFT.Fragments.Fragment): ending fragment.
      steps (int): the number of steps to take between A and B.
      extrapolation_steps (int): optionally, we can extrapolate a number of
        steps beyond B on the same trajectory.

    Returns:
      (list): a list of fragments interpolating between A and B including
      A and B.
    """
    from numpy import matrix, array
    from BigDFT.wahba import interpolate_points
    from copy import deepcopy

    pos1 = matrix([at.get_position() for at in A])
    pos2 = matrix([at.get_position() for at in B])

    point_list = interpolate_points(pos1, pos2, steps, extrapolation_steps)

    frag_list = []
    for i in range(0, steps+extrapolation_steps+2):
        new_frag = deepcopy(A)
        for j in range(0, len(new_frag)):
            new_frag[j].set_position([float(x)
                                      for x in array(point_list[i][j])[0]])
        frag_list.append(new_frag)

    return frag_list


class Fragment(MutableSequence):
    """
    A fragment is a list of atoms in a system. Fragment might have quantities
    associated to it, like its electrostatic multipoles (charge, dipole, etc.)
    and also geometrical information (center of mass, principla axis etc.). A
    Fragment might also be rototranslated and combined with other moieteies to
    form a :class:`BigDFT.Fragments.System`.

    Args:
      atomlist (list): list of atomic dictionaries defining the fragment
      xyzfile (XYZReader): an XYZ file to read from.
      posinp (dict): the posinp style dictionary from a logfile/input file.

    .. todo::
       Define and describe if this API is also suitable for solid-state
       fragments

    """

    def __init__(self, atomlist=None, xyzfile=None, posinp=None, astruct=None):
        from BigDFT.Atom import Atom
        self.atoms = []

        # insert atoms.
        if atomlist:
            for atom in atomlist:
                self.append(Atom(atom))
        elif xyzfile:
            with xyzfile:
                for line in xyzfile:
                    self.append(Atom(line))
        elif posinp:
            units = posinp.get('units', 'angstroem')
            for atom in posinp['positions']:
                self.append(Atom(atom, units=units))
        elif astruct:
            units = astruct.get('units', 'angstroem')
            rshift = astruct.get('Rigid Shift Applied (AU)', [0.0, 0.0, 0.0])
            for atom in astruct["positions"]:
                self.append(Atom(atom, units=units))
            self.translate([-1.0*x for x in rshift])

        # Values
        self.purity_indicator = None
        self.q0 = None
        self.q1 = None
        self.q2 = None
        self.frozen = None
        self.conmat = None

    def __len__(self):
        return len(self.atoms)

    def __delitem__(self, index):
        self.atoms.__delitem__(index)

    def insert(self, index, value):
        from BigDFT.Atom import Atom
        self.atoms.insert(index, Atom(value))

    def __setitem__(self, index, value):
        from BigDFT.Atom import Atom
        self.atoms.__setitem__(index, Atom(value))

    def __getitem__(self, index):
        # If they ask for only one atom, then we return it as an atom.
        # but if it's a range we return a ``Fragment`` with those atoms in it.
        if isinstance(index, slice):
            return Fragment(atomlist=self.atoms.__getitem__(index))
        else:
            return self.atoms.__getitem__(index)

    def __add__(self, other):
        from copy import deepcopy
        rval = deepcopy(self)
        rval += other
        return rval

    @property
    def centroid(self):
        """
        The center of a fragment.
        """
        from numpy import mean, ravel
        pos = [at.get_position() for at in self]
        return ravel(mean(pos, axis=0))

    def center_of_charge(self, zion):
        """
        The charge center which depends both on the position and net charge
        of each atom.
        """
        from numpy import array
        cc = array([0.0, 0.0, 0.0])
        qtot = 0.0
        for at in self:
            netcharge = at.q0
            zcharge = zion[at.sym]
            elcharge = zcharge - netcharge
            cc += elcharge * array(at.get_position())
            qtot += elcharge
        return cc / qtot

    def d0(self, center=None):
        """
        Fragment dipole, calculated only from the atomic charges.

        Args:
          center (list): the center of charge of the fragment.
            If this is not present, the centroid is used.
        """
        from numpy import zeros, array
        # one might added a treatment for non-neutral fragments
        # but if the center of charge is used the d0 value is zero
        if center is not None:
            cxyz = center
        else:
            cxyz = self.centroid

        d0 = zeros(3)
        found = False
        for at in self:
            if at.q0 is not None:
                found = True
                d0 += at.q0 * (array(at.get_position()) - array(cxyz))

        if found:
            return d0
        else:
            return None

    def d1(self, center=None):
        """
        Fragment dipole including the atomic dipoles.

        Args:
          center (list): the center of charge of the fragment.
            If this is not present, the centroid is used.
        """
        from numpy import zeros
        d1 = zeros(3)
        dtot = self.d0(center)
        if dtot is None:
            return dtot

        found = False
        for at in self:
            if at.q1 is not None:
                found = True
                d1 += at.q1

        if found:
            return d1 + dtot
        else:
            return None
        pass

    def ellipsoid(self, center=0.0):
        """
        Todo: define the ellipsoid.
        """
        import numpy as np
        Imat = np.mat(np.zeros(9).reshape(3, 3))
        for at in self:
            rxyz = np.array(at.get_position()) - center
            Imat[0, 0] += rxyz[0]**2  # rxyz[1]**2+rxyz[2]**2
            Imat[1, 1] += rxyz[1]**2  # rxyz[0]**2+rxyz[2]**2
            Imat[2, 2] += rxyz[2]**2  # rxyz[1]**2+rxyz[0]**2
            Imat[0, 1] += rxyz[1] * rxyz[0]
            Imat[1, 0] += rxyz[1] * rxyz[0]
            Imat[0, 2] += rxyz[2] * rxyz[0]
            Imat[2, 0] += rxyz[2] * rxyz[0]
            Imat[1, 2] += rxyz[2] * rxyz[1]
            Imat[2, 1] += rxyz[2] * rxyz[1]
        return Imat

    def fragment_transformation(self, frag2):
        """
        Returns the transformation among fragments if it exists.

        Args:
          frag2 (BigDFT.Fragments.Fragment): the fragment to transform between.

        Returns:
          (BigDFT.Fragments.RotoTranslation) : the transformation matrix.
        """
        from numpy import mat
        pos1 = mat(self.centroid)
        pos2 = mat(frag2.centroid)
        return RotoTranslation(pos1, pos2)

    def get_external_potential(self, units="bohr", charge_offset=False):
        """
        Transform the fragment information into a dictionary ready to be
        put as an external potential.

        Args:
          units (str): the units of the external potential.

        Returns:
          (dict): a dictionary describing the external potential for use in
          an input file.
        """
        pot = [at.get_external_potential(units) for at in self]

        return pot

    def get_net_force(self):
        """
        Returns the net force on a fragment in Ha/Bohr.

        Returns:
          (list) Three values which describe the net force.
        """
        from numpy import array
        ret_val = array([0.0, 0.0, 0.0])
        for at in self:
            ret_val += array(at.get_force())
        return [float(x) for x in ret_val]

    def line_up(self):
        """
        Align the principal axis of inertia of the fragments along the
        coordinate axis. Also shift the fragment such as its centroid is zero.
        """
        from numpy.linalg import eig
        Shift = Translation(self.centroid)
        Shift.invert()
        self.transform(Shift)
        # now the centroid is zero
        Imat = self.ellipsoid()
        w, v = eig(Imat)
        Redress = Rotation(v.T)
        self.transform(Redress)
        # now the principal axis of inertia are on the coordinate axis

    @property
    def qcharge(self):
        """
        The net charge on a fragment.
        """
        from BigDFT.Atom import nzion
        netcharge = self.q0[0]
        for at in self:
            if "nzion" in at:
                zcharge = at.qcharge["nzion"]
            else:
                zcharge = nzion(at.sym)
            netcharge += zcharge
        return netcharge

    def rotate(self, x=None, y=None, z=None, units="radians"):
        """
        Rotate the fragment.

        Args:
          x (float): angle to rotate on the x axis.
          y (float): angle to rotate on the y axis.
          z (float): angle to rotate on the z axis.
          units (str): either radians or degrees.
        """
        from math import cos, sin, pi
        from numpy import mat, identity
        from copy import deepcopy

        # Deal with the units.
        if units == "degrees":
            if x:
                xval = x * pi / 180
            if y:
                yval = y * pi / 180
            if z:
                zval = z * pi / 180
        elif units == "radians":
            xval = x
            yval = y
            zval = z
        else:
            raise ValueError("Units must be degrees or radians")

        # Translate back to the origin.
        centroid = self.centroid
        self.translate(-1.0 * centroid)

        # Build the rotation
        rot = identity(3)
        if x:
            rx = mat([
                [1.0,        0.0,       0.0],
                [0.0,        cos(xval), -sin(xval)],
                [0.0,        sin(xval),  cos(xval)]
            ])
            rot = rx.dot(rot)
        if y:
            ry = mat([
                [cos(yval), 0.0,        sin(yval)],
                [0.0,        1.0,        0.0],
                [-sin(yval), 0.0,        cos(yval)]
            ])
            rot = ry.dot(rot)
        if z:
            rz = mat([
                [cos(zval), -sin(zval), 0.0],
                [sin(zval),  cos(zval), 0.0],
                [0.0,        0.0,       1.0]
            ])
            rot = rz.dot(rot)

        # Rotate
        rt = Rotation(rot)
        rself = rt.dot(self)
        for i in range(0, len(self)):
            self[i] = deepcopy(rself[i])

        # Translate back
        self.translate(centroid)

    def translate(self, vec):
        """
        Translate the fragment along the vector provided.

        Args:
          vec (list): a list of x, y, z values describing the translation (AU).
        """
        from copy import deepcopy
        rt = Translation(vec)

        trans = rt.dot(self)
        for i in range(0, len(self)):
            self[i] = deepcopy(trans[i])


def system_from_log(log, fragmentation=None):
    """
    This function returns a :class:`~BigDFT.Fragment.System` class out of a
    logfile. If the logfile contains information about fragmentation and atomic
    multipoles, then the system is created accordingly.
    Otherwise, the fragmentation scheme is determined by the fragmentation
    variable.

    Args:
       log (Logfile): the logfile of the QM run. In general must have been done
           with Linear Scaling formalism.
       fragmentation (str): the scheme to be used for the fragmentation in the
           case if not provided internally by the logfile.
           The possible values are ``atomic`` and ``full``, in which case the
           system as as many fragments as the number of atoms, or only one
           fragment, respectively.
    Returns:
        (BigDFT.Fragments.System) The instance of the class containing
        fragments.
    """
    name = log.log.get('run_name', 'FULL') + ':0'

    full_system = System()
    if "posinp" in log.log:
        posinp = log.log['posinp']
        full_system[name] = Fragment(posinp=posinp)
    else:
        full_system[name] = Fragment(astruct=log.astruct)

    # provide the atomic information on the system
    if hasattr(log, 'electrostatic_multipoles'):
        full_system.set_atom_multipoles(log)
        if hasattr(log, 'forces'):
            full_system.set_atom_forces(log)

    # now we may defragment the system according to the provided scheme
    if fragmentation == 'full':
        return full_system
    elif fragmentation == 'atomic' or 'posinp' not in log.log:
        atomic_system = System()
        for iat, at in enumerate(full_system[name]):
            atomic_system['ATOM:' + str(iat)] = Fragment([at])
        return atomic_system
    else:
        posinp = log.log['posinp']
        frag_dict = {}
        for iat, tupl in enumerate(zip(posinp['positions'],
                                       full_system[name])):
            at, obj = tupl
            fragid = at.get('frag', 'ATOM:' + str(iat))
            if isinstance(fragid, list):
                fragid = str(fragid[0]) + ':' + str(fragid[1])
            if fragid not in frag_dict:
                frag_dict[fragid] = [obj]
            else:
                frag_dict[fragid].append(obj)
        frag_system = System()
        for fragid in frag_dict:
            frag_system[fragid] = Fragment(frag_dict[fragid])
        return frag_system


class System(MutableMapping):
    """
    A system is defined as a named collection of fragments. You can manipulate
    a system as if it were a standard python dictionary, however it also has
    helper routines for performing operations on the full system.
    """

    def __init__(self, *args, **kwargs):
        self.store = dict()
        self.update(dict(*args, **kwargs))
        self.conmat = None

    def dict(self):
        """
        Convert to a dictionary.
        """
        return self.store

    def __getitem__(self, key):
        return self.store[self.__keytransform__(key)]

    def __setitem__(self, key, value):
        self.store[self.__keytransform__(key)] = value

    def __delitem__(self, key):
        del self.store[self.__keytransform__(key)]

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def __keytransform__(self, key):
        return key

    @property
    def centroid(self):
        """
        Center of mass of the system
        """
        from numpy import mean
        return mean([frag.centroid for frag in self.values()], axis=0)

    @property
    def central_fragment(self):
        """
        Returns the fragment whose center of mass is closest to the centroid

        Returns:
          (str): the name of the fragment.
          (Fragment): the fragment object
        """
        import numpy as np
        CMs = [frag.centroid for frag in self.values()]
        idx = np.argmin([np.dot(dd, dd.T) for dd in (CMs - self.centroid)])
        return list(self.keys())[idx], list(self.values())[idx]

    def get_external_potential(self, units="bohr", charge_offset=False):
        """
        Transform the system information into a dictionary ready to be
        put as an external potential.

        Args:
          units (str): the units of the external potential.
          charge_offset (bool): by default the external potential ignores the
            counter charge from the protons. Setting this to true adds the
            positive charge to the potential.
        """
        ret_dict = {"units": units}
        ret_dict["values"] = []
        for frag in self.values():
            ret_dict["values"].extend(
                frag.get_external_potential(units, charge_offset))
        ret_dict["global monopole"] = sum(
            x["q0"][0] for x in ret_dict["values"])
        return ret_dict

    def get_k_nearest_fragments(self, target, k, cutoff=None):
        """
        Given a fragment id in a system, this computes the nearest fragment.

        Args:
          target (str): the fragment to find the nearest neighbor of.
          k (int): the number of fragments to look for.
          cutoff (float): will only return fragments with a certain cutoff.

        Returns:
          (lists): the ids of the nearest fragments.
        """
        from scipy.spatial import KDTree

        # Setup the KD Tree for distance lookup
        poslist = []
        frag_lookup = []
        for fragid, frag in self.items():
            if fragid == target:
                continue
            for at in frag:
                poslist.append(at.get_position())
                frag_lookup.append(fragid)
        tree = KDTree(poslist)

        # Find the nearest fragments with a query of the tree.
        targetpost = [x.get_position() for x in self[target]]
        if cutoff is not None:
            ndist, nearest = tree.query(targetpost, k=k,
                                        distance_upper_bound=cutoff)
        else:
            ndist, nearest = tree.query(targetpost, k=k)

        # We now have the nearest atom to each atom in this fragment.
        # Next we combine this information and extract the closest
        # fragments.

        if k == 1:
            ndist = [ndist]
            nearest = [nearest]

        distdict = {}
        for i in range(0, len(nearest)):
            for idx, dist in zip(nearest[i], ndist[i]):
                try:
                    fragidx = frag_lookup[idx]
                except IndexError:
                    # kdtree returns invalid indices if it can't find enough
                    # points.
                    continue
                if fragidx not in distdict:
                    distdict[fragidx] = dist
                elif distdict[fragidx] < dist:
                    distdict[fragidx] = dist

        # Extract the k smallest values.
        minlist = []
        for i in range(0, k):
            if len(distdict) == 0:
                break
            key = min(distdict, key=distdict.get)
            minlist.append(key)
            del distdict[key]

        return minlist

    def get_nearest_fragment(self, target):
        """
        Given a fragment id in a system, this computes the nearest fragment.

        Args:
          target (str): the fragment to find the nearest neighbor of.

        Returns:
          (str): the id of the nearest fragment.
        """
        return self.get_k_nearest_fragments(target, k=1)[0]

    def get_net_force(self):
        """
        Returns the net force on a system in Ha/Bohr.

        Returns:
          (list): Three values which describe the net force.
        """
        from numpy import array
        ret_val = array([0.0, 0.0, 0.0])
        for frag in self.values():
            ret_val += array(at.get_net_force())
        return [float(x) for x in ret_val]

    def get_posinp(self, units='angstroem'):
        """
        Provide the dictionary which has to be passed to the ``posinp`` value
        of the :meth:`run` method of  the
        :class:`~BigDFT.Calculators.SystemCalculator` class instance.

        Args:
           units (str): The units of the file. May be "angstroem" or "bohr".
        """
        pos = []
        for fragid, frag in self.items():
            for at in frag:
                atdict = {at.sym: at.get_position(units)}
                atdict["frag"] = list(GetFragTuple(fragid))
                if frag.frozen:
                    atdict["Frozen"] = frag.frozen
                pos.append(atdict)
        return {'units': units, 'positions': pos}

    @property
    def q0(self):
        """
        Provides the global monopole of the system given as a sum of the
        monopoles of the atoms.
        """
        if len(self) == 0:
            return None
        return [sum(filter(None, [frag.q0[0] for frag in self.values()]))]

    @property
    def qcharge(self):
        """
        The total qcharge of a system.
        """
        return sum([frag.qcharge for frag in self.values()])

    def rename_fragments(self):
        """
        This procedure automatically names the fragments in a system.

        Returns:
          (System): the same system, with the automatic naming scheme.
        """
        rnsys = System()
        for i, fragid in enumerate(self):
            rnsys["FRAG:"+str(i)] = self[fragid]
        return rnsys

    def set_atom_multipoles(self, logfile, correct_charge=True):
        """
        After a run is completed, we have a set of multipoles defined on
        each atom. This routine will set those values on to each atom
        in the system.

        Args:
          logfile (Logfiles.Logfile): logfile with the multipole values.
          correct_charge (bool): currently there is an inconsistency in
            terms of gross charge, and this corrects it.
        """
        mp = logfile.electrostatic_multipoles
        for pole in mp["values"]:
            pole["units"] = mp["units"]
        lookup = self.compute_matching(mp["values"])

        # Assign
        for fragid, frag in self.items():
            for i, at in enumerate(frag):
                idx = lookup[fragid][i]
                if idx >= 0:
                    at.set_multipole(mp["values"][idx], correct_charge)

    def set_atom_forces(self, logfile):
        """
        After a run is completed, we have the forces on each atom in the
        logfile. This routine will set those values to each atom in this sytem.

        Args:
          logfile (Logfiles.Logfile): logfile with the forces.
        """
        # We will use the multipoles to help figure out which force value
        # is associated with which atom
        mp = logfile.electrostatic_multipoles
        lookup = self.compute_matching(mp["values"])

        # Assign forces
        forces = logfile.forces
        for fragid, frag in self.items():
            for i, at in enumerate(frag):
                idx = lookup[fragid][i]
                if idx >= 0:
                    at.set_force(list(forces[idx].values())[0])

    def write_fragfile(self, filename, logfile):
        """
        Write out the file needed as input to the fragment analysis routine.

        Args:
          filename (str): name of the file to write to.
          logfile (Logfiles.Logfile): the log file this calculation is based
            on (to ensure a matching order of atoms).
        """
        from yaml import dump

        # Extract the indices
        mp = logfile.electrostatic_multipoles
        lookup = self.compute_matching(mp["values"])
        outlist = []
        for fragid, frag in self.items():
            outlist.append([])
            for i, at in enumerate(frag):
                idx = lookup[fragid][i]
                outlist[-1].append(idx + 1)

        # Write
        with open(filename, "w") as ofile:
            dump(outlist, ofile)

    def write_pdb(self, filename):
        """
        Write out a system to a pdb file.

        Args:
          filename (str): the file to write to.
        """
        # Put all the data into this string.
        outstr = ""

        idx = 1
        lookup = {}
        for fragid, frag in self.items():
            for i, at in enumerate(frag):
                pos = [str("{:.3f}".format(x))
                       for x in at.get_position("angstroem")]
                fragtuple = GetFragTuple(fragid)

                line = list(" " * 80)
                line[0:6] = "HETATM"  # HETATM
                line[7:11] = str(idx).rjust(4)  # SERIAL NUMBER
                line[12:16] = at.sym.ljust(4)  # ATOM NAME
                line[16:17] = " "  # ALTERNATIVE LOCATION INDICATOR
                line[17:20] = fragtuple[0][:3].ljust(3)  # RESIDUE NAME
                line[21:22] = "A"  # CHAIN IDENTIFIER
                line[22:26] = fragtuple[1].rjust(3)  # RESIDUE SEQUENCE NUMBER
                line[26:27] = " "  # CODE FOR INSERTION OF RESIDUES
                line[30:38] = pos[0].rjust(8)  # X COORDINATE
                line[38:46] = pos[1].rjust(8)  # Y COORDINATE
                line[46:54] = pos[2].rjust(8)  # Z COORDINATE
                line[54:60] = "      "  # OCCUPANCY
                line[60:66] = "      "  # TEMPERATURE
                line[72:76] = "    "  # SEGMENT IDENTIFIER
                line[76:78] = at.sym.rjust(2)  # ELEMENT SYMBOL
                line[78:80] = "  "  # CHARGE
                outstr += line + "\n"

                # Keep track of the indexes
                lookup[(fragid, i)] = idx
                idx = idx + 1

        # Write the connectivity information
        if self.conmat is not None:
            for fragid, frag in self.items():
                for i, at in enumerate(frag):
                    connections = self.conmat[fragid][i]
                    line = list(" " * 80)
                    line[0:6] = "CONECT"
                    # SERIAL NUMBER
                    line[7:11] = str(lookup[(fragid, i)]).rjust(4)
                    if len(connections) > 0:  # BOND SERIAL NUMBERS
                        line[12:16] = str(lookup[connections[0]]).rjust(4)
                    if len(connections) > 1:
                        line[17:21] = str(lookup[connections[1]]).rjust(4)
                    if len(connections) > 2:
                        line[22:26] = str(lookup[connections[2]]).rjust(4)
                    if len(connections) > 3:
                        line[27:31] = str(lookup[connections[3]]).rjust(4)
                    outstr += line + "\n"

        # Finally, write out to file.
        with open(filename, "w") as ofile:
            ofile.write(outstr)

    def compute_matching(self, atlist, shift=None):
        """
        Frequently we are passed a list of atom like objects from which we
        need to extract data and assign it to a system. However, a system
        can potentially store those atoms in any order, and may not have
        the same set of atoms. This helper routine creates a mapping between
        this list view, to the dictionary view of the system class.

        Args:
          atlist (list): a list of atom like objects.
          shift (list): if the positions in atlist are shifted by some constant
            vector you can specify that here.

        Returns:
          (dict): a mapping from a system to indices in the atom list. If
            an atom is not in the list, an index value of -1 is assigned.
        """
        from BigDFT.Atom import Atom
        from numpy import array
        from scipy.spatial import KDTree

        # Convert everything to pure positions to avoid overhead.
        poslist = [array(Atom(x).get_position("bohr")) for x in atlist]
        if shift is not None:
            poslist = [x - array(shift) for x in poslist]
        tree = KDTree(poslist)

        # Seach for the mapping values
        mapping = {}
        for fragid, frag in self.items():
            mapping[fragid] = []
            for at in frag:
                atpos = array(at.get_position("bohr"))
                ndist, nearest = tree.query(atpos)
                mapping[fragid].append(nearest)

        return mapping


if __name__ == "__main__":
    from BigDFT.XYZ import XYZReader, XYZWriter
    from os.path import join
    from os import system
    from copy import deepcopy

    safe_print("Read in an xyz file and build from a list.")
    atom_list = []
    with XYZReader(join("Database", "XYZs", "SiO.xyz")) as reader:
        for at in reader:
            atom_list.append(at)
    frag1 = Fragment(atomlist=atom_list)
    for at in frag1:
        safe_print(at.sym, at.get_position())
    safe_print("Centroid", frag1.centroid)
    safe_print()

    safe_print("Build from an xyz file directory.")
    reader = XYZReader(join("Database", "XYZs", "Si4.xyz"))
    frag2 = Fragment(xyzfile=reader)
    for at in frag2:
        safe_print(at.sym, at.get_position())
    safe_print()

    safe_print("We can combine two fragments with +=")
    frag3 = deepcopy(frag1)
    frag3 += frag2
    for at in frag3:
        safe_print(at.sym, at.get_position())
    safe_print("Length of frag3", len(frag3))
    safe_print()

    safe_print("Since we can iterate easily, we can also write easily.")
    with XYZWriter("test.xyz", len(frag3), "angstroem") as writer:
        for at in frag3:
            writer.write(at)
    system("cat test.xyz")
    safe_print()

    safe_print("We can also extract using the indices")
    safe_print(dict(frag3[0]))
    sub_frag = frag3[1:3]
    for at in sub_frag:
        safe_print(dict(at))
    safe_print()

    safe_print("Now we move on to testing the system class.")
    safe_print("We might first begin in the easiest way.")
    sys = System(frag1=frag1, frag2=frag2)
    for at in sys["frag1"]:
        safe_print(dict(at))
    for at in sys["frag2"]:
        safe_print(dict(at))
    safe_print()

    safe_print("What if we want to combine two fragments together?")
    sys["frag1"] += sys.pop("frag2")
    for at in sys["frag1"]:
        safe_print(dict(at))
    safe_print("frag2" in sys)
    safe_print()

    safe_print("What if I want to split a fragment by atom indices?")
    temp_frag = sys.pop("frag1")
    sys["frag1"], sys["frag2"] = temp_frag[0:3], temp_frag[3:]
    for at in sys["frag1"]:
        safe_print(dict(at))
    for at in sys["frag2"]:
        safe_print(dict(at))
    safe_print()

    safe_print("Construct a system from an XYZ file.")
    fname = join("Database", "XYZs", "BH2.xyz")
    sys2 = System(frag1=Fragment(xyzfile=XYZReader(fname)))

    safe_print("Split it to fragments")
    sys2["frag1"], sys2["frag2"] = sys2["frag1"][0:1], sys2["frag1"][1:]

    safe_print("And write to file")
    with XYZWriter("test.xyz", len(frag3), "angstroem") as writer:
        for fragid, frag in sys2.items():
            for at in frag:
                writer.write(at)
    system("cat test.xyz")



import yaml
from futile.Utils import write
    
#function which removes from a set of lines the yaml_fields contained in the to_remove list
def clean_logfile(logfile_lines,to_remove):
  """Remove yaml fields from a list of lines.

  Removes from a set of lines the yaml_fields contained in the to_remove list.

  Arguments:
      logfile_lines (list): list of the lines of the logfile. Generated from a file by e.g. :py:meth:`~io.IOBase.readlines`.
      to_remove (list): list of keys to remove from logfile_lines

  Returns:
      list of lines where the removed keys have as values the `"<folded>"` string
  """
  line_rev=logfile_lines #list of the lines of the logfile
  #loop in the reversed from (such as to parse by blocks)
  extra_lines=20 #internal variable to be customized
  line_rev.reverse()
  #clean the log
  cleaned_logfile=[]
  removed=[]
  #for line in line_rev: #line_iter:
  while len(line_rev) >0:
    line=line_rev.pop()
    to_print=line
    #check if the line contains interesting information
    for remove_it in to_remove :
      stream_list=[]
      #line without comments
      valid_line=line.split('#')[0]
      spaces='nospace'
      #control that the string between the key and the semicolon is only spaces
      if remove_it in valid_line and ":" in valid_line:
        #print "here",remove_it,remove_it in valid_line and ":" in valid_line,valid_line
        starting_point=valid_line.find(remove_it)
        tmp_buf=valid_line[:starting_point]
        #find the closest comma to the staring point, if exists
        tmp_buf=tmp_buf[::-1]
        starting_comma=tmp_buf.find(',')
        if starting_comma <0: st=0
        tmp_buf=tmp_buf[st:]
        tmp_buf=tmp_buf[::-1]
        tmp_buf=tmp_buf.strip(' ')
        #print "there",tmp_buf,'starting',starting_point,len(tmp_buf)
        valid_line= valid_line[starting_point+len(remove_it):]
        spaces= valid_line[1:valid_line.find(':')]
        #if remove_it+':' in line.split('#')[0]:
      if len(spaces.strip(' ')) == 0 and len(tmp_buf)==0: #this means that the key has been found
         #creates a new Yaml document starting from the line
         #treat the rest of the line following the key to be removed
         header=''.join(line.split(':')[1:])
         header=header.rstrip()+'\n'
         #eliminate the anchor
         header=header.lstrip(' ')
         header=header.lstrip('*') 
         if len(header) > 0 :
            stream_list.append(header)
         #part to be printed, updated
         to_print = line.split(':')[0] + ": <folded> \n"

         #then check when the mapping will end:
         while True:
            #create a stream with extra_lines block           
            for i in range(0,min(extra_lines,len(line_rev))):
               stream_list.append(line_rev.pop())
            #create a stream to be parsed
            stream=''.join(stream_list)
            #then parse the stream until the last valid position has been found
            try:
              for i in yaml.parse(stream,Loader=yaml.CLoader):
                endpos=i.end_mark.index
            except Exception(e):
              #  print 'error',str(e),stream
              #convert back the valid stream into a list
              #if needed the stream can be loaded into a document
              item_list=stream[:endpos].split('\n')
              #if lengths are different there is no need to add lines
              if len(item_list) != len(stream_list):
                #last line might be shorter, therefore treat it separately
                last_line=item_list.pop()
                #purge the stream
                for item in item_list:
                  stream_list.remove(item+'\n')
                #extract the remaining line which should be compared with the last one
                strip_size=len(last_line.rstrip())
                if strip_size > 0:
                  first_line=stream_list.pop(0)[strip_size:]
                  if '*' in first_line or '&' in first_line:
                    first_line='' #eliminate anchors
                else:
                  first_line=''
                #then put the rest in the line to be treated
                to_print.rstrip('\n')
                to_print += first_line+'\n'
                # the item has been found
                break
         stream_list.reverse()
         #put back the unused part in the document
         line_rev.extend(stream_list)
         # mark that the key has been removed
         if (remove_it not in removed):
           removed.append(remove_it)
           write('removed: ',remove_it)
  # then print out the line 
    cleaned_logfile.append(to_print)

  # check that everything has been removed, at least once
  if (set(removed) != set(to_remove)):
    write('WARNING, not all the requested items have been removed!')
    write('To_remove : ',to_remove)
    write('removed   : ',removed)
    write('Difference: ',list(set(to_remove) - set(removed) ))
  return cleaned_logfile


def load(file=None,stream=None,doc_lists=True,safe_mode=False):
    """Encapsulate the loading of yaml documents.
    
    Provides a dictionary, or a list of dictionaries, which 
    represents the structure of the stream to be loaded.
    It also wraps the yaml loader to perform a optimized parsing when the
    `minloader` of PyYaml 3.13 is available.
    This wrapper ensures to extract from the stream the maximum possible information
    by choosing the best loader available.

    Arguments:
        file (str): path of the yaml-compliant file containing the stream to be loaded

        stream (str): the stream to load, overrides the ``file`` argument if present

        doc_lists (bool): if True, ensures that the results is always in a form 
           of lists of documents, even in the case of a single doc
           When False, the return type is either a dictionary or a generator according
           to the specifications of yaml.load and yaml.load_all respectively.

        safe_mode (bool): When true, in the case of multiple documents 
           in the stream, it loads the document one after another. 
           This is useful to avoid losing of all the document list 
           in the case when one of the document is 
           not yaml compliant, like in the case of a broken logfile. 
           It may works only when the separation of the 
           documents is indicated by the usual syntax ``"---\\n"`` 
           (i.e. no yaml tags between documents)

    Returns:
        * a list of dictionaries, if ``doc_lists`` is set to ``True``;
        * a dictionary, if the stream or the file contains a single yaml document;
        * a generator if the parsed stream is made of multiple documents *and* ``safe_mode`` = ``False``;
        * a list of dictionaries if the stream is made of multiple documents and ``safe_mode`` is ``True``.
    """
    #Detect None otherwise a doc == '' gives an error
    strm=stream if stream != None else open(file,'r').read()
    #choose the loader
    try:
        ldr=yaml.MinLoader
    except:
        try:
            ldr=yaml.CLoader
        except:
            ldr=yaml.Loader

    #load the documents
    ld=[]
    try:
        ld=yaml.load(strm,Loader=ldr)
        if doc_lists: ld=[ld]
    except Exception as e:
        if safe_mode:
            ld = []
            documents = [v for v in strm.split('---\n') if len(v) > 0]
            for i,raw_doc in enumerate(documents):
                try:
                    ld.append(yaml.load(raw_doc,Loader=ldr))
                except Exception as f:
                    write('Document',i,'of stream NOT loaded, error:',f)
        else:
            ld=yaml.load_all(strm,Loader=ldr)
            if doc_lists: ld=[l for l in ld]
    return ld

def dump(data,filename=None,raw=False,tar=False):
    """Encapsulate the dumping of dictionaries.
    
    This function is useful to dump a dictionary in yaml or json-compliant form.
    This may be used as an alternative to the usual :py:meth:`yaml.dump <https://pyyaml.org/wiki/PyYAMLDocumentation>` method,
    especially when the dictionary to be dump'ed is heavy.
    No particular attention is paid in human readability of the output.
    The dumped information can then be parsed either from json or yaml interpreter.

    Arguments:
       data (dict,list): the information to be dumped
       filename (str): path of the file in which the information will be stored.
          If absent, the information is written on :py:func:`sys.stdout`.
       raw (bool): if ``True`` the output is in json-style, otherwise it is pre-processed by :py:meth:yaml.dump, 
          but ``None`` is passed to ``default_flow_style``.
       tar (bool): if ``True`` the filename is assumed to be a compressed tarfile. 
          The :py:mod:`tarfile` module is used to create and append information.
    """
    todump=str(data) if raw else yaml.safe_dump(data,default_flow_style=None)
    if filename:
        if tar:
            import tarfile
            from cStringIO import StringIO
            import time
            f=tarfile.TarInfo(filename)
            f.size=len(str(todump))
            f.mtime=time.time()
            tar.addfile(f,StringIO(str(todump)))
        else:
            f=open(filename,'w')
            f.write(todump)
    else:
        import sys
        sys.stdout.write(todump)

class YamlDB(dict):
    """
    Yaml Database, read from a file or a stream

    :param str file: take the database from a file
    :param str stream: associate the stream as the value of the dictionary
    :param bool singledoc: guarantees that the provided stream can only contain one document
    :param list ignore: A list of keys that will not be considered in the loading.
            Useful to parse log logfiles in less time

    * YamlDB(file) -> Database from a file
    * YamlDB(stream) -> Database from a stream
    * YamlDB() -> new empty dictionary
    * YamlDB(mapping) -> new dictionary initialized from a mapping object's (key, value) pairs
    * YamlDB(key, stream=stream) -> new YamlDB with unparsed stream written as the value of the key.
    * YamlDB(iterable) -> new dictionary initialized as if via:
           d = {}
           for k, v in iterable:
               d[k] = v
    * YamlDB(**kwargs) -> new dictionary initialized with the name=value pairs
      in the keyword argument list.  For example:  YamlDB(one=1, two=2)
    """
    def __init__(self,*args,**kwargs): #stream=None,singledoc=False):
        """Extract the database information"""
        from futile.Utils import kw_pop
        newkw,self.singledoc=kw_pop('singledoc',False,**kwargs)
        newkw,self.file=kw_pop('file',None,**newkw)
        newkw,self.stream=kw_pop('stream',None,**newkw)
        newkw,self.ignore=kw_pop('ignore',None,**newkw)
        listargs=list(args)
        if len(listargs)>0:
            if type(listargs[0])==type('string'):
                import os
                tmp=listargs.pop(0)
                if self.file is None and os.path.isfile(tmp):
                    self.file=tmp
                elif self.stream is None:
                    self.stream=tmp
        #suppose that the stream is a file
        if self.file is not None:
            sl=open(self.file,'r')
            prestream=''.join(sl.readlines())
            sl.close()
        elif self.stream is not None:
            prestream=self.stream
        else:
            super(YamlDB,self).__init__(*listargs,**newkw)
            return
        if self.singledoc:
            self.stream=prestream
        else:
            #process stream, only one document
            startpos=0
            i=0
            for endpos in self._doc_finder(prestream):
                if i==1: self.docs=[YamlDB(stream=doctmp,singledoc=True,ignore=self.ignore)]
                doctmp=prestream[startpos:endpos]
                if i>=1: self.docs.append(YamlDB(stream=doctmp,singledoc=True,ignore=self.ignore))
                startpos=endpos
                i+=1
            if i==1: self.stream=doctmp
        if self.stream is not None: 
            dd=self._load()
            if dd: super(YamlDB,self).update(dd)
    def __len__(self):
        if hasattr(self,'docs'):
            return len(self.docs)
        else:
            return 0
    def __getitem__(self,key):
        if isinstance(key,int):
            if hasattr(self,'docs'):
                return self.docs[key]
        else:
            return self.get(key)
    def documents(self):
        "Generator over the whole documents"
        if hasattr(self,'docs'):
            for d in self.docs:
                yield dict(d)
        else:
            yield dict(self)
    def _doc_finder(self,stream):
        #first find the stream of the document start
        startpos=0
        import sys
        write('Loading...')
        while startpos < len(stream):
            try:
                startpos+=self._event_finder(stream[startpos:],yaml.events.DocumentEndEvent)
            except:
                startpos=len(stream)
            write(int((100.0*startpos)/len(stream)),'%')
            yield startpos
    def _load(self):
        #clean the logfile if needed
        if self.ignore is not None:
            listlog=[ a+'\n' for a in self.stream.split('\n')]
            cleanedlog=clean_logfile(listlog,self.ignore)
            stream='\n'.join(cleanedlog)
        else:
            stream=self.stream
        try:
            return yaml.load(stream,Loader=yaml.CLoader)
        except Exception(e):
            #here we might put a bigger 
            return None
    def _event_finder(self,present,event,end=False):
        for i in yaml.parse(present,Loader=yaml.CLoader):
            if isinstance(i,event):
                if end:
                    key=i.value
                    startpos=i.start_mark.index
                    endpos=startpos+self._find_endblock(present[startpos:])
                    substream=present[i.end_mark.index:endpos]
                    return key,substream,endpos
                else:
                    startpos=i.end_mark.index
                    return startpos
        #in the case of absent event
        if end:
            write( 'here',present,'end')
            return 'nothing','',len(present)
        else:
            return len(present)  
    def _find_endblock(self,stream):
        """Find the end of the block which is yaml-compliant"""
        endpos=0
        try:
            for i in yaml.parse(stream,Loader=yaml.CLoader):
                endpos=i.end_mark.index
        except yaml.YAMLError(e):
            #stop at the last carriage return
            endpos=e.problem_mark.index
            endpos=stream.rfind('\n',0,endpos)
        return endpos



"""Handle the input variable specifications.

This module is the python complement to the specification of the input variables of a 
file as provided by the :f:mod:`f_input_file` module. It uses the same syntax as defined there 
and make possible the interplay between a python-based serialization of the input dictionary 
and the construction of 

"""

class InputVariable():
    """
    Define a input variable of a library.
    Such object can be initialized and inspected from the dictionary used by futile 
    and then initialized according to the provided specification.
    """
    def __init__(self,name,spec):
        """
        Define a specification of the input variable, according to the conventions
        employed in the :f:mod:`f_input_file` module.
        
        Args:
           name (str): the name of the input variable
           spec (dict): dictionary describing the specification of the variables

        Todo: 
           Extensively specify the conventions somewhere

        """
        from futile.Utils import kw_pop
        if 'default' not in spec or 'COMMENT' not in spec:
            raise ValueError("Invalid specification of the variable '"+name+"'")
        self.name=name
        self.spec,self.valid_range=kw_pop('RANGE',None,**spec)
        self.spec,self.valid_values=kw_pop('EXCLUSIVE',None,**self.spec)
        self.spec,desc=kw_pop('DESCRIPTION','None',**self.spec)
        self.spec,self.profile_from=kw_pop('PROFILE_FROM',None,**self.spec)
        self.spec,self.condition=kw_pop('CONDITION',None,**self.spec)
        self.impose_profile_to=[]
        self._val_from_profile('default')
        self.__doc__=self.name+":\n"+self.spec.pop("COMMENT")+"\nDescription:\n"+desc
    def _val(self,val):
        if self.valid_range and             (val < self.valid_range[0] or val > self.valid_range[1]):
                raise ValueError("Variable '"+ str(val)+ "' not in allowed range " + str(self.valid_range))
        if self.valid_values and (val not in self.valid_values):
            raise ValueError("Variable '" + str(val) + "' not in allowed values " + str(self.valid_values))
        self.value=val
    def _val_from_profile(self,profile):
        self.profile=profile
        val=self.spec[profile]
        if type(val) == str and val in self.spec: val=self.spec[val]
        self._val(val)
        #if the profile chosen is also present in the input variables that 
        #depends from that then impose it
        for name,var in self.impose_profile_to:
            if self.profile in var.spec: var.set(self.profile)
    def is_valid(self):
        if self.condition is not None and "VARIABLE" in self.condition:   
            #raise ValueError("Variable '"+self.name+"' requires to be bound with master_key '"+\
            #                 self.condition["MASTER_KEY"]+"';\n use the set_master_variable method")
            var=self.condition["VARIABLE"]
            when=self.condition.get("WHEN")
            if when and (var.profile not in when and var.value not in when):
                return False #raise ValueError("Variable '"+self.name+"' cannot be set as master key '"+\ self.condition["MASTER_KEY"]+"' does not have the allowed values")
            when_not=self.condition.get("WHEN_NOT")
            if when_not and (var.profile in when_not or var.value in when_not):
                return False #raise ValueError("Variable '"+self.name+"' cannot be set as master key '"+self.condition["MASTER_KEY"]+"' has incompatible values")
        return True
    def _val_from_user(self,val):
        self.profile='__USER__'
        #search if some profile corresponds to a value
        for profile,value in self.spec.items():
            if value == val:
                self._val_from_profile(profile)
                return
        self._val(val)
    def set(self,val):
        """Set the value of the input variable"""
        if val in self.spec:
            self._val_from_profile(profile=val)
        else:
            self._val_from_user(val)
    def set_dependent_variable(self,var):
        """Set the dependent variable from which impose the profile"""
        self.impose_profile_to.append((var.name,var))
    def set_master_variable(self,var):
        """Set the variable which is activated when the present has suitable values or profiles"""
        if self.condition: 
            if var.name != self.condition["MASTER_KEY"]:
                raise ValueError("Variable '"+self.name+"' requires to be bound with '"+                                self.condition["MASTER_KEY"]+"', whereas '"+                                var.name+"' has been provided")
            self.condition["VARIABLE"]=var
    def __repr__(self):
        if self.profile != '__USER__':
            return self.profile
        else:
            return str(self.value)




"""
This module is useful to process a logfile of BigDFT run, in yaml format.
It also provides some tools to extract typical informations about the run,
like the energy, the eigenvalues and so on.
"""

# This module needs: os, yaml, futile, matplotlib, numpy, BZ, DoS

EVAL = "eval"
SETUP = "let"
INITIALIZATION = "globals"

PATH = 'path'
PRINT = 'print'
GLOBAL = 'global'
FLOAT_SCALAR = 'scalar'

PRE_POST = [EVAL, SETUP, INITIALIZATION]

# Builtin paths to define the search paths
BUILTIN = {
    'number_of_orbitals': {PATH: [['Total Number of Orbitals']],
                           PRINT: "Total Number of Orbitals", GLOBAL: True},
    'posinp_file': {PATH: [['posinp', 'properties', 'source', ]],
                    PRINT: "source:", GLOBAL: True},
    'XC_parameter': {PATH: [['dft', 'ixc'], ['DFT parameters:', 'XC ID:']],
                     PRINT: "ixc:", GLOBAL: True, FLOAT_SCALAR: True},
    'grid_spacing': {PATH: [["dft", "hgrids"]],
                     PRINT: "hgrids:", GLOBAL: True},
    'spin_polarization': {PATH: [["dft", "nspin"]],
                          PRINT: "nspin:", GLOBAL: True},
    'total_magn_moment': {PATH: [["dft", "mpol"]],
                          PRINT: "mpol:", GLOBAL: True},
    'system_charge': {PATH: [["dft", "qcharge"]],
                      PRINT: "qcharge:", GLOBAL: True},
    'rmult': {PATH: [["dft", "rmult"]],
              PRINT: "rmult:", GLOBAL: True},
    # 'up_elec'::{PATH: [["occupation:","K point 1:","up:","Orbital \d+"]],
    #       PRINT: "Orbital \d+", GLOBAL: True},
    'astruct': {PATH: [['Atomic structure']]},
    'data_directory': {PATH: [['Data Writing directory']]},
    'dipole': {PATH: [['Electric Dipole Moment (AU)', 'P vector']],
               PRINT: "Dipole (AU)"},
    'electrostatic_multipoles': {PATH: [['Multipole coefficients']]},
    'energy': {PATH: [["Last Iteration", "FKS"], ["Last Iteration", "EKS"],
                      ["Energy (Hartree)"]],
               PRINT: "Energy", GLOBAL: False},
    'evals': {PATH: [["Complete list of energy eigenvalues"],
                     ["Ground State Optimization", -1, "Orbitals"],
                     ["Ground State Optimization", -1,
                      "Hamiltonian Optimization", -1, "Subspace Optimization",
                      "Orbitals"]]},
    'fermi_level': {PATH: [["Ground State Optimization", -1, "Fermi Energy"],
                           ["Ground State Optimization", -1,
                            "Hamiltonian Optimization", -1,
                            "Subspace Optimization", "Fermi Energy"]],
                    PRINT: True, GLOBAL: False},
    'forcemax': {PATH: [["Geometry", "FORCES norm(Ha/Bohr)", "maxval"],
                        ['Clean forces norm (Ha/Bohr)', 'maxval']],
                 PRINT: "Max val of Forces"},
    'forcemax_cv': {PATH: [['geopt', 'forcemax']],
                    PRINT: 'Convergence criterion on forces',
                    GLOBAL: True, FLOAT_SCALAR: True},
    'force_fluct': {PATH: [["Geometry", "FORCES norm(Ha/Bohr)", "fluct"]],
                    PRINT: "Threshold fluctuation of Forces"},
    'forces': {PATH: [['Atomic Forces (Ha/Bohr)']]},
    'gnrm_cv': {PATH: [["dft", "gnrm_cv"]],
                PRINT: "Convergence criterion on Wfn. Residue", GLOBAL: True},
    'kpts': {PATH: [["K points"]],
             PRINT: False, GLOBAL: True},
    'kpt_mesh': {PATH: [['kpt', 'ngkpt']], PRINT: True, GLOBAL: True},
    'magnetization': {PATH: [["Ground State Optimization", -1,
                              "Total magnetization"],
                             ["Ground State Optimization", -1,
                              "Hamiltonian Optimization", -1,
                              "Subspace Optimization", "Total magnetization"]],
                      PRINT: "Total magnetization of the system"},
    'memory_run': {PATH: [
      ['Accumulated memory requirements during principal run stages (MiB.KiB)']
    ]},
    'memory_quantities': {PATH: [
      ['Memory requirements for principal quantities (MiB.KiB)']]},
    'memory_peak': {PATH: [['Estimated Memory Peak (MB)']]},
    'nat': {PATH: [['Atomic System Properties', 'Number of atoms']],
            PRINT: "Number of Atoms", GLOBAL: True},
    'pressure': {PATH: [['Pressure', 'GPa']], PRINT: True},
    'sdos': {PATH: [['SDos files']], GLOBAL: True},
    'support_functions': {PATH: [["Gross support functions moments",
                                  'Multipole coefficients', 'values']]},
    'symmetry': {PATH: [['Atomic System Properties', 'Space group']],
                 PRINT: "Symmetry group", GLOBAL: True}}


def get_logs(files):
    """
    Return a list of loaded logfiles from files, which is a list
    of paths leading to logfiles.

    :param files: List of filenames indicating the logfiles
    :returns: List of Logfile instances associated to filename
    """
    from futile import YamlIO
    logs = []
    for filename in files:
        logs += YamlIO.load(filename, doc_lists=True, safe_mode=True)
    return logs


def floatify(scalar):
    """
    Useful to make float from strings compatible from fortran

    Args:
       scalar (str, float): When string representing a float that might be
         given in fortran notation, otherwise it might be a floating point
    Returns:
       float. The value associated to scalar as a floating point number

    Example:
       >>> # this would be the same with "1.e-4" or with 0.0001
       >>> floatify('1.d-4')
       1.e-4

    """
    if isinstance(scalar, str):
        return float(scalar.replace('d', 'e').replace('D', 'E'))
    else:
        return scalar


# This is a tentative function written to extract information from the runs
def document_quantities(doc, to_extract):
    """
    Extract information from the runs.

    .. warning::
        This routine was designed for the previous parse_log.py script and it
        is here only for backward compatibility purposes.
    """
    analysis = {}
    for quantity in to_extract:
        if quantity in PRE_POST:
            continue
        # follow the levels indicated to find the quantity
        field = to_extract[quantity]
        if not isinstance(field, list) and not isinstance(field, dict) \
                and field in BUILTIN:
            paths = BUILTIN[field][PATH]
        else:
            paths = [field]
        # now try to find the first of the different alternatives
        for path in paths:
            # print path,BUILTIN,BUILTIN.keys(),field in BUILTIN,field
            value = doc
            for key in path:
                # as soon as there is a problem the quantity is null
                try:
                    value = value[key]
                except (KeyError, TypeError):
                    value = None
                    break
            if value is not None:
                break
        analysis[quantity] = value
    return analysis


def perform_operations(variables, ops, debug=False):
    """
    Perform operations given by 'ops'.
    'variables' is a dictionary of variables i.e. key=value.

    .. warning::
       This routine was designed for the previous parse_log.py script and it is
       here only for backward compatibility purposes.
    """
# glstr=''
# if globs is not None:
# for var in globs:
#            glstr+= "global "+var+"\n"
# if debug: print '###Global Strings: \n',glstr
# first evaluate the given variables
    for key in variables:
        command = key+"="+str(variables[key])
        if debug:
            print(command)
        exec(command)
        # then evaluate the given expression
    if debug:
        print(ops)
    # exec(glstr+ops, globals(), locals())
    exec(ops, globals(), locals())


def process_logfiles(files, instructions, debug=False):
    """
    Process the logfiles in files with the dictionary 'instructions'.

    .. warning::
       This routine was designed for the previous parse_log.py script and it is
       here only for backward compatibility purposes.
    """
    import sys
    glstr = 'global __LAST_FILE__ \n'
    glstr += '__LAST_FILE__='+str(len(files))+'\n'
    if INITIALIZATION in instructions:
        for var in instructions[INITIALIZATION]:
            glstr += "global "+var+"\n"
            glstr += var + " = " + str(instructions[INITIALIZATION][var])+"\n"
            # exec var +" = "+ str(instructions[INITIALIZATION][var])
    exec(glstr, globals(), locals())
    for f in files:
        sys.stderr.write("#########processing "+f+"\n")
        datas = get_logs([f])
        for doc in datas:
            doc_res = document_quantities(doc, instructions)
            # print doc_res,instructions
            if EVAL in instructions:
                perform_operations(doc_res, instructions[EVAL], debug=debug)


def find_iterations(log):
    """
    Identify the different block of the iterations of the wavefunctions
    optimization.

    .. todo::
       Should be generalized and checked for mixing calculation and O(N)
       logfiles

    :param log: logfile load
    :type log: dictionary
    :returns: wavefunction residue per iterations, per each subspace
      diagonalization
    :rtype: numpy array of rank two
    """
    import numpy
    for itrp in log['Ground State Optimization']:
        rpnrm = []
        for itsp in itrp['Hamiltonian Optimization']:
            gnrm_sp = []
            for it in \
                    itsp['Subspace Optimization']['Wavefunctions Iterations']:
                if 'gnrm' in it:
                    gnrm_sp.append(it['gnrm'])
            rpnrm.append(numpy.array(gnrm_sp))
    rpnrm = numpy.array(rpnrm)
    return rpnrm


def plot_wfn_convergence(wfn_it, gnrm_cv, label=None):
    """
    Plot the convergence of the wavefunction coming from the find_iterations
    function. Cumulates the plot in matplotlib.pyplot module

    :param wfn_it: list coming from :func:`find_iterations`
    :param gnrm_cv: convergence criterion for the residue of the wfn_it list
    :param label: label for the given plot
    """
    import matplotlib.pyplot as plt
    import numpy
    plt.semilogy(numpy.ravel(wfn_it), label=label)
    plt.legend(loc="upper right")
    plt.axhline(gnrm_cv, color='k', linestyle='--')
    it = 0
    for itrp in wfn_it:
        it += len(itrp)
        plt.axvline(it, color='k', linestyle='--')


class Logfile():
    """
    Import a Logfile from a filename in yaml format, a list of filenames,
    an archive (compressed tar file), a dictionary or a list of dictionaries.

    :param args: logfile names to be parsed
    :type args: strings
    :param kwargs: keyword arguments

       * archive: name of the archive from which retrieve the logfiles
       * member: name of the logfile within the archive. If absent, all the
           files of the archive will be considered as args
       * label: the label of the logfile instance
       * dictionary: parsed logfile given as a dictionary, serialization of the
           yaml logfile

    :Example:
       >>> l = Logfile('one.yaml','two.yaml')
       >>> l = Logfile(archive='calc.tgz')
       >>> l = Logfile(archive='calc.tgz',member='one.yaml')
       >>> l = Logfile(dictionary=dict1)
       >>> l = Logfile(dictionary=[dict1, dict2])

    .. todo::
       Document the automatically generated attributes, perhaps via an inner
       function in futile python module

    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the class
        """
        import os
        dicts = []
        # Read the dictionary kwargs
        arch = kwargs.get("archive")
        member = kwargs.get("member")
        label = kwargs.get("label")
        dictionary = kwargs.get("dictionary")
        if arch:
            # An archive is detected
            import tarfile
            from futile import YamlIO
            tar = tarfile.open(arch)
            members = [tar.getmember(member)] if member else tar.getmembers()
            # print members
            for memb in members:
                f = tar.extractfile(memb)
                dicts.append(YamlIO.load(stream=f.read()))
                # Add the label (name of the file)
                # dicts[-1]['label'] = memb.name
            srcdir = os.path.dirname(arch)
            label = label if label is not None else arch
        elif dictionary:
            # Read the dictionary or a list of dictionaries or from a generator
            # Need to return a list
            dicts = [dictionary] if isinstance(dictionary, dict) else [
                d for d in dictionary]
            srcdir = ''
            label = label if label is not None else 'dict'
        elif args:
            # Read the list of files (member replaces load_only...)
            dicts = get_logs(args)
            label = label if label is not None else args[0]
            srcdir = os.path.dirname(args[0])
        #: Label of the Logfile instance
        self.label = label
        #: Absolute path of the directory of logfile
        self.srcdir = os.path.abspath('.' if srcdir == '' else srcdir)
        if not dicts:
            raise ValueError("No log information provided.")
        # So we have a list of a dictionary or a list of dictionaries
        # Initialize the logfile with the first document
        self._initialize_class(dicts[0])
        #
        if len(dicts) > 1:
            # first initialize the instances with the previous logfile such as
            # to provide the correct information (we should however decide what
            # to do if some run did not converged)
            self._instances = []
            for i, d in enumerate(dicts):
                # label=d.get('label','log'+str(i))
                label = 'log'+str(i)
                dtmp = dicts[0]
                # Warning: recursive call!!
                instance = Logfile(dictionary=dtmp, label=label)
                # now update the instance with the other value
                instance._initialize_class(d)
                self._instances.append(instance)
            # then we should find the best values for the dictionary
            print('Found', len(self._instances), 'different runs')
            import numpy
            # Initialize the class with the dictionary corresponding to the
            # lower value of the energy
            ens = [(l.energy if hasattr(l, 'energy') else 1.e100)
                   for l in self._instances]
            #: Position in the logfile items of the run associated to lower
            #  energy
            self.reference_log = numpy.argmin(ens)
            # print 'Energies',ens
            self._initialize_class(dicts[self.reference_log])
    #

    def __getitem__(self, index):
        if hasattr(self, '_instances'):
            return self._instances[index]
        else:
            # print('index not available')
            raise ValueError(
                'This instance of Logfile has no multiple instances')
    #

    def __str__(self):
        """Display short information about the logfile"""
        return self._print_information()
    #

    def __len__(self):
        if hasattr(self, '_instances'):
            return len(self._instances)
        else:
            return 0  # single point run
    #

    def _initialize_class(self, d):
        import numpy
        # : dictionary of the logfile (serialization of yaml format)
        self.log = d
        # here we should initialize different instances of the logfile class
        # again
        sublog = document_quantities(self.log, {val: val for val in BUILTIN})
        for att, val in sublog.items():
            if val is not None:
                val_tmp = floatify(val) if BUILTIN[att].get(
                    FLOAT_SCALAR) else val
                setattr(self, att, val_tmp)
            elif hasattr(self, att) and not BUILTIN[att].get(GLOBAL):
                delattr(self, att)
        # then postprocess the particular cases
        if not hasattr(self, 'fermi_level') and hasattr(self, 'evals'):
            self._fermi_level_from_evals(self.evals)

        if hasattr(self, 'kpts'):
            #: Number of k-points, present only if meaningful
            self.nkpt = len(self.kpts)
            if hasattr(self, 'evals'):
                self.evals = self._get_bz(self.evals, self.kpts)
            if hasattr(self, 'forces') and hasattr(self, 'astruct'):
                self.astruct.update({'forces': self.forces})
                delattr(self, 'forces')
        elif hasattr(self, 'evals'):
            from BigDFT import BZ
            #: Eigenvalues of the run, represented as a
            #  :class:`BigDFT.BZ.BandArray` class instance
            self.evals = [BZ.BandArray(self.evals), ]
        if hasattr(self, 'sdos'):
            import os
            # load the different sdos files
            sd = []
            for f in self.sdos:
                try:
                    data = numpy.loadtxt(os.path.join(self.srcdir, f))
                except IOError:
                    data = None
                if data is not None:
                    xs = []
                    ba = [[], []]
                    for line in data:
                        xs.append(line[0])
                        ss = self._sdos_line_to_orbitals(line)
                        for ispin in [0, 1]:
                            ba[ispin].append(ss[ispin])
                    sd.append({'coord': xs, 'dos': ba})
                else:
                    sd.append(None)
            #: Spatial density of states, when available
            self.sdos = sd
        # memory attributes
        self.memory = {}
        for key in ['memory_run', 'memory_quantities', 'memory_peak']:
            if hasattr(self, key):
                title = BUILTIN[key][PATH][0][0]
                self.memory[title] = getattr(self, key)
                if key != 'memory_peak':
                    delattr(self, key)
    #

    def _fermi_level_from_evals(self, evals):
        import numpy
        # this works when the representation of the evals is only with
        # occupied states
        # write('evals',self.evals)
        fl = None
        fref = None
        for iorb, ev in enumerate(evals):
            e = ev.get('e')
            if e is not None:
                fref = ev['f'] if iorb == 0 else fref
                fl = e
                if ev['f'] < 0.5*fref:
                    break
            e = ev.get('e_occ', ev.get('e_occupied'))
            if e is not None:
                fl = e if not isinstance(
                    e, list) else numpy.max(numpy.array(e))
            e = ev.get('e_vrt', ev.get('e_virt'))
            if e is not None:
                break
        #: Chemical potential of the system
        self.fermi_level = fl
    #

    def _sdos_line_to_orbitals_old(self, sorbs):
        from BigDFT import BZ
        evals = []
        iorb = 1
        # renorm=len(xs)
        # iterate on k-points
        kpts = self.kpts if hasattr(self, 'kpts') else [
            {'Rc': [0.0, 0.0, 0.0], 'Wgt':1.0}]
        for i, kp in enumerate(kpts):
            ev = []
            # iterate on the subspaces of the kpoint
            for ispin, norb in enumerate(self.evals[0].info):
                for iorbk in range(norb):
                    # renorm postponed
                    ev.append({'e': sorbs[iorb+iorbk],
                               's': 1-2*ispin, 'k': i+1})
                    # ev.append({'e':np.sum([ so[iorb+iorbk] for so in sd]),
                    #            's':1-2*ispin,'k':i+1})
                iorb += norb
            evals.append(BZ.BandArray(
                ev, ikpt=i+1, kpt=kp['Rc'], kwgt=kp['Wgt']))
        return evals
    #

    def _sdos_line_to_orbitals(self, sorbs):
        import numpy as np
        iorb = 1
        sdos = [[], []]
        for ikpt, band in enumerate(self.evals):
            sdoskpt = [[], []]
            for ispin, norb in enumerate(band.info):
                if norb == 0:
                    continue
                for i in range(norb):
                    val = sorbs[iorb]
                    iorb += 1
                    sdoskpt[ispin].append(val)
                sdos[ispin].append(np.array(sdoskpt[ispin]))
        return sdos
    #

    def _get_bz(self, ev, kpts):
        """Get the Brillouin Zone."""
        evals = []
        from BigDFT import BZ
        for i, kp in enumerate(kpts):
            evals.append(BZ.BandArray(
                ev, ikpt=i+1, kpt=kp['Rc'], kwgt=kp['Wgt']))
        return evals
    #

    def get_dos(self, label=None, npts=2500):
        """
        Get the density of states from the logfile.

        :param label: id of the density of states.
        :type label: string
        :param npts: number of points of the DoS curve
        :type npts: int
        :returns: Instance of the DoS class
        :rtype: :class:`BigDFT.DoS.DoS`
        """
        from BigDFT import DoS
        # reload(DoS)
        lbl = self.label if label is None else label
        sdos = self.sdos if hasattr(self, 'sdos') else None
        return DoS.DoS(bandarrays=self.evals, label=lbl, units='AU',
                       fermi_level=self.fermi_level, npts=npts, sdos=sdos)
    #

    def get_brillouin_zone(self):
        """
        Return an instance of the BrillouinZone class, useful for band
        structure.
        :returns: Brillouin Zone of the logfile
        :rtype: :class:`BigDFT.BZ.BrillouinZone`
        """
        from BigDFT import BZ
        if self.nkpt == 1:
            print('WARNING: Brillouin Zone plot cannot be defined properly'
                  ' with only one k-point')
            # raise
        mesh = self.kpt_mesh  # : K-points grid
        if isinstance(mesh, int):
            mesh = [mesh, ]*3
        if self.astruct['cell'][1] == float('inf'):
            mesh[1] = 1
        return BZ.BrillouinZone(self.astruct, mesh, self.evals,
                                self.fermi_level)
    #

    def wfn_plot(self):
        """
        Plot the wavefunction convergence.
        :Example:
           >>> tt=Logfile('log-with-wfn-optimization.yaml',label='a label')
           >>> tt.wfn_plot()
        """
        wfn_it = find_iterations(self.log)
        plot_wfn_convergence(wfn_it, self.gnrm_cv, label=self.label)
    #

    def geopt_plot(self):
        """
        For a set of logfiles construct the convergence plot if available.
        Plot the Maximum value of the forces against the difference between
        the minimum value of the energy and the energy of the iteration.
        Also an errorbar is given indicating the noise on the forces for a
        given point. Show the plot as per plt.show() with matplotlib.pyplots as
        plt

        :Example:
           >>> tt=Logfile('log-with-geometry-optimization.yaml')
           >>> tt.geopt_plot()
        """
        energies = []
        forces = []
        ferr = []
        if not hasattr(self, '_instances'):
            print('ERROR: No geopt plot possible, single point run')
            return
        for l in self._instances:
            if hasattr(l, 'forcemax') and hasattr(l, 'energy'):
                forces.append(l.forcemax)
                energies.append(l.energy-self.energy)
                ferr.append(0.0 if not hasattr(l, 'force_fluct') else (
                    self.force_fluct if hasattr(self, 'force_fluct') else 0.0))
        if len(forces) > 1:
            import matplotlib.pyplot as plt
            plt.errorbar(energies, forces, yerr=ferr,
                         fmt='.-', label=self.label)
            plt.legend(loc='upper right')
            plt.loglog()
            plt.xlabel('Energy - min(Energy)')
            plt.ylabel('Forcemax')
            if hasattr(self, 'forcemax_cv'):
                plt.axhline(self.forcemax_cv, color='k', linestyle='--')
            plt.show()
        else:
            print('No plot necessary, less than two points found')
    #
    #

    def _print_information(self):
        """Display short information about the logfile (used by str)."""
        import yaml
        # summary=[{'Atom types':
        #          numpy.unique([at.keys()[0] for at in
        #                       self.astruct['positions']]).tolist()},
        #          {'cell':
        #           self.astruct.get('cell', 'Free BC')}]
        summary = [{'Atom types':
                    self.log['Atomic System Properties']['Types of atoms']},
                   {'cell':
                    self.astruct.get('cell', 'Free BC')}]
        # normal printouts in the document, according to definition
        for field in BUILTIN:
            name = BUILTIN[field].get(PRINT)
            if name:
                name = field
            if not name or not hasattr(self, field):
                continue
            summary.append({name: getattr(self, field)})
        if hasattr(self, 'evals'):
            nspin = self.log['dft']['nspin']
            if nspin == 4:
                nspin = 1
            cmt = (' per k-point' if hasattr(self, 'kpts') else '')
            summary.append(
                {'No. of KS orbitals'+cmt: self.evals[0].info[0:nspin]})
        return yaml.dump(summary, default_flow_style=False)


if __name__ == "__main__":
    # Create a logfile: should give an error
    # (ValueError: No log information provided.)
    lf = Logfile()



"""
Basic wrapping workchain on a BigDFT computation.

"""
from aiida import orm
from aiida.common.extendeddicts import AttributeDict
from aiida.engine import BaseRestartWorkChain, process_handler, while_, ExitCode
from aiida.plugins import CalculationFactory, DataFactory

from aiida.engine.processes.workchains.utils import process_handler, ProcessHandlerReport

from futile import YamlIO

Dict = DataFactory('dict') 

RemoteData = DataFactory('remote') 
BigDFTCalculation = CalculationFactory('bigdft')

class BigDFTBaseWorkChain(BaseRestartWorkChain):

    _process_class = BigDFTCalculation

    @classmethod
    def define(cls, spec):
        super(BigDFTBaseWorkChain, cls).define(spec)
        spec.input('show_warnings', valid_type=orm.Bool,
                   default=lambda: orm.Bool(True),
                   help='turn the warnings on/off.')
        spec.input('run_opts', valid_type=Dict,
                    required=False,
                    help='metadata')
        spec.expose_inputs(BigDFTCalculation, exclude=('metadata',))

        spec.outline(
            cls.setup,
            while_(cls.should_run_process)(
                cls.run_process,
                cls.inspect_process,
            ),
            cls.results,
        )
        spec.expose_outputs(BigDFTCalculation)
        #this one needs to be optional to avoid being checked wrongly by the restartworkchain
        spec.exit_code(100, 'ERROR_INPUT',
                       message='BigDFT input error')
        spec.exit_code(200, 'ERROR_RUNTIME',
                       message='BigDFT runtime error')

    @process_handler(priority=600)
    def check_debug_output(self, calculation):
        repo = calculation.outputs.retrieved._repository._get_base_folder()
        try:
            repo.get_abs_path('debug', check_existence=True)
        except OSError:
            return
        debug_folder = repo.get_subfolder('debug')
        # debug folder exists, error probably happened.
        if "jobname" in self.ctx.inputs.metadata.options:
            jobname = self.ctx.inputs.metadata.options.jobname
        else:
            jobname = 'BigDFT job'
        posout_list = debug_folder.get_content_list(pattern="bigdft-err*")
        for filename in posout_list:
            log = YamlIO.load(debug_folder.get_abs_path(filename))
            err = log[0].get('BIGDFT_INPUT_VARIABLES_ERROR')
            info = log[0].get('Additional Info')
            if err is not None:
                self.report('{}<{}> input error : {} Id: {}.\n\
                            Additional Information : {}'.
                            format(jobname, calculation.pk,
                                   err['Message'], err['Id']))
                if info is not None:
                    self.report('Additional Info :', info)
                return ProcessHandlerReport(True, ExitCode(100))
            err = log[0].get('BIGDFT_RUNTIME_ERROR')
            if err is not None:
                self.report('{}<{}> runtime error : {} Id: {}'.
                            format(jobname, calculation.pk,
                                   err['Message'], err['Id']))
                if info is not None:
                    self.report('Additional Info :', info)
                return ProcessHandlerReport(True, ExitCode(200))

    @process_handler(priority=590)
    def check_warnings(self, calculation):
        if calculation.is_finished_ok and self.inputs.show_warnings:
            logfile=calculation.outputs.bigdft_logfile.logfile
            if(isinstance(logfile, list)):
                warnings = logfile[-1].get('WARNINGS')
            else:
                warnings = logfile.get('WARNINGS')
            if warnings is not None:
                self.report('Warnings were found :')
                for warn in warnings:
                    self.report(warn)

    @process_handler(priority=500)
    def finish(self, calculation):
        if calculation.is_finished_ok:
            if "jobname" in self.ctx.inputs.metadata.options:
                jobname = self.ctx.inputs.metadata.options.jobname
            else:
                jobname = 'BigDFT job'
            self.report('{}<{}> completed successfully'.
                        format(jobname, calculation.pk))
            self.ctx.restart_calc = calculation
            self.ctx.is_finished = True

    def setup(self):
        super().setup()
        self.ctx.inputs = AttributeDict(self.exposed_inputs(BigDFTCalculation))
        if self.inputs.get('run_opts') is not None:
            self.ctx.inputs.metadata = AttributeDict(self.inputs.run_opts.get_dict())
        else:
            self.ctx.inputs.metadata = {}





""" Tests for command line interface.

"""
from __future__ import print_function
from __future__ import absolute_import

from aiida.manage.fixtures import PluginTestCase


class TestDataCli(PluginTestCase):
    def setUp(self):
        from click.testing import CliRunner
        from aiida.plugins import DataFactory

        BigDFTParameters = DataFactory('bigdft')
        self.parameters = BigDFTParameters({'ignore-case': True})
        self.parameters.store()
        self.runner = CliRunner()

    def test_data_BigDFT_list(self):
        """Test 'verdi data bigdft list'

        Tests that it can be reached and that it lists the node we have set up.
        """
        from aiida_bigdft.cli import list_

        result = self.runner.invoke(list_, catch_exceptions=False)
        self.assertIn(str(self.parameters.pk), result.output)

    def test_data_BigDFT_export(self):
        """Test 'verdi data bigdft export'

        Tests that it can be reached and that it shows the contents of the node
        we have set up.
        """
        from aiida_bigdft.cli import export

        result = self.runner.invoke(export, [str(self.parameters.pk)],
                                    catch_exceptions=False)
        self.assertIn('ignore-case', result.output)



# This module needs: numpy, matplotlib, scipy, ase, spglib
import numpy
from futile.Utils import write as safe_print


def get_ev(ev, keys=None, ikpt=1):
    """Get the correct list of the energies for this eigenvalue."""
    res = False
    if keys is None:
        ener = ev.get('e')
        spin = ev.get('s')
        kpt = ev.get('k')
        if not kpt and ikpt == 1:
            kpt = True
        elif kpt and kpt != ikpt:
            kpt = False
        if ener and (spin == 1 or not spin):
            if kpt:
                res = [ener]
        elif ener and spin == -1:
            if kpt:
                res = [None, ener]
    else:
        for k in keys:
            if k in ev:
                res = ev[k]
                if not isinstance(res, []):
                    res = [res]
                break
    return res


def astruct_to_cell(astruct):
    """
    Convert the astruct information as parsed from the module Logfiles into
    the cell structure as needed from spglib
    """
    import numpy
    celltmp = [a if a != float('inf') else 1.0 for a in astruct['cell']]
    lattice = numpy.diag(celltmp)
    pos = [[a/b if b != float('inf') else 0.0
            for a, b in zip(list(at.values())[0], celltmp)]
           for at in astruct['positions']]
    atoms = [list(at.keys())[0] for at in astruct['positions']]
    ianames, iatype = numpy.unique(atoms, return_inverse=True)
    return (lattice, pos, iatype)


class BandArray(numpy.ndarray):
    """
    Defines the array of data for one band. It is a dictionary which contains
    a numpy array for both spin channels.
    """
    def __new__(cls, *args, **kwargs):  # logdata,ikpt=0,kpt=(0.0,0.0,0.0):
        "Takes the data from the logfile and convert it"
        datain = kwargs.get("data", None)
        if datain is not None:
            evs = datain
            shape0 = len(evs)
            norbs = list(map(len, evs))
            if len(norbs) == 1:
                norbs = [norbs[0], 0]
        else:
            evs = [[], []]
            logdata = kwargs.get("logdata", args[0])
            ikpt = kwargs.get("ikpt", 1 if len(args) < 2 else args[1])
            # Ugly patch to detect proper k point in Davidson...
            prev_vrt = True
            cur_ikpt = 0
            # End of ugly patch
            for ev in logdata:
                occ = get_ev(ev, ['e_occ', 'e_occupied'], ikpt=ikpt)
                vrt = get_ev(ev, ['e_vrt', 'e_virt'], ikpt=ikpt)
                eigen = occ or vrt
                # Ugly patch to detect proper k point in Davidson...
                if occ and prev_vrt:
                    cur_ikpt += 1
                prev_vrt = vrt
                if eigen and cur_ikpt != ikpt:
                    continue
                # End of ugly patch
                if not eigen:
                    eigen = get_ev(ev, ikpt=ikpt)
                if not eigen:
                    continue
                for i, e in enumerate(eigen):
                    if e:
                        evs[i].append(e)
            shape0 = 2 if len(evs[1]) > 0 else 1
            norbs = list(map(len, evs))
        data = numpy.ndarray.__new__(cls, shape=(shape0, max(norbs)),
                                     dtype=numpy.float)
        data.fill(numpy.nan)
        data[0, :len(evs[0])] = evs[0]
        if norbs[1] > 0:
            data[1, :len(evs[1])] = evs[1]
        data.info = norbs  # (map(len,evs))
        return data

    def __init__(self, *args, **kwargs):
        ikpt = kwargs.get("ikpt", 1 if len(args) < 2 else args[1])
        kpt = kwargs.get("kpt", (0., 0., 0.) if len(args) < 3 else args[2])
        kwgt = kwargs.get('kwgt', 1.0)
        self.set_kpt(ikpt, kpt, kwgt)

    def set_kpt(self, ikpt, kpt, kwgt=1.0):
        if not isinstance(ikpt, int):
            raise TypeError('ikpt should be a integer')
        if len(kpt) != 3:
            raise TypeError('kpt should be a object of len 3')
        self.ikpt = ikpt
        self.kpt = kpt
        self.kwgt = kwgt

    def __add__(self, b):
        if hasattr(b, 'kpt') and (b.kpt != self.kpt):
            raise ValueError('cannot sum BandArray with different kpoints')
        if hasattr(b, 'kwgt') and (b.kwgt != self.kwgt):
            raise ValueError('cannot sum BandArray with different kweights')
        c = super(type(self), self).__add__(b)
        return BandArray(data=c, ikpt=self.ikpt, kpt=self.kpt, kwgt=self.kwgt)


class BZPath():
    """
    Defines a set of points which are associated to a path in the reduced
    Brillouin Zone.
    """

    def __init__(self, lattice, path, special_points, npts=50):
        import ase.dft.kpoints as ase
        self.special_points = special_points
        path_tmp = []
        self.symbols = []
        # construct the path
        for p in path:
            if isinstance(p, str):
                # then this is a special point
                path_tmp.append(self.special_points[p])
                self.symbols.append(p.replace('G', '$\\Gamma$'))
            else:
                path_tmp.append(list(p.values())[0])
                self.symbols.append(list(p.keys())[0])
        self.path, self.xaxis, self.xlabel = ase.get_bandpath(
            path_tmp, lattice, npts)


class BrillouinZone():
    def __init__(self, astruct, mesh, evals, fermi_energy):
        import spglib
        import numpy
        cell = astruct_to_cell(astruct)
        self.lattice = cell[0]
        # celltmp=[ a if a!=float('inf') else 1.0 for a in astruct['cell']]
        # self.lattice=numpy.diag(celltmp)
        # print 'lattice',self.lattice
        # pos=[[ a/b if b!=float('inf') else 0.0
        #        for a,b in zip(at.values()[0], celltmp)]
        #      for at in astruct['positions']]
        # atoms=[ at.keys()[0] for at in astruct['positions']]
        # ianames,iatype=numpy.unique(atoms,return_inverse=True) #[1,]*4+[2,]*4
        # we should write a function for the iatype
        # print 'iatype', iatype
        # cell=(self.lattice,pos,iatype)
        safe_print('spacegroup', spglib.get_spacegroup(cell, symprec=1e-5))
        # then define the pathes and special points
        import ase.dft.kpoints as ase
        # we should adapt the 'cubic'
        cell_tmp = astruct['cell']
        # print 'cell',
        #        cell_tmp,numpy.allclose(cell_tmp,[cell_tmp[0],]*len(cell_tmp))
        if numpy.allclose(cell_tmp, [cell_tmp[0], ]*len(cell_tmp)):
            lattice_string = 'cubic'
        else:
            lattice_string = 'orthorhombic'
        safe_print('Lattice found:', lattice_string)
        self.special_points = ase.get_special_points(
            lattice_string, self.lattice, eps=0.0001)
        self.special_paths = ase.parse_path_string(
            ase.special_paths[lattice_string])
        self.fermi_energy = fermi_energy
        # dataset = spglib.get_symmetry_dataset(cell, symprec=1e-3)
        # print dataset
        # the shift has also to be put if present
        mapping, grid = spglib.get_ir_reciprocal_mesh(
            mesh, cell, is_shift=[0, 0, 0])
        lookup = []
        for ikpt in numpy.unique(mapping):
            ltmp = []
            for ind, (m, g) in enumerate(zip(mapping, grid)):
                if m == ikpt:
                    ltmp.append((g, ind))
            lookup.append(ltmp)
        safe_print('irreductible k-points', len(lookup))
        # print 'mapping',mapping
        # print 'grid',len(grid),numpy.max(grid)
        coords = numpy.array(grid, dtype=numpy.float)/mesh
        # print 'coords',coords
        # print 'shape',coords.shape
        # print grid #[ x+mesh[0]*y+mesh[0]*mesh[1]*z for x,y,z in grid]
        # brillouin zone
        kp = numpy.array([k.kpt for k in evals])
        ourkpt = numpy.rint(kp*(numpy.array(mesh))).astype(int)
        # print ourkpt
        bz = numpy.ndarray((coords.shape[0], evals[0].size), dtype=float)
        # print bz
        # shift = (numpy.array(mesh)-1)/2
        # print 'shift',shift
        for ik in lookup:
            irrk = None
            for orbs, bzk in zip(evals, ourkpt):
                for (kt, ind) in ik:
                    if (bzk == kt).all():
                        irrk = orbs
                        # print 'hello',orbs.kpt,kt
                        break
                if irrk is not None:
                    break
            if irrk is None:
                safe_print('error in ik', ik)
                safe_print('our', ourkpt)
                safe_print('spglib', grid)
                safe_print('mapping', mapping)
            for (kt, ind) in ik:
                # r=kt+shift
                # ind=numpy.argwhere([(g==kt).all() for g in grid])
                # print 'ik',kt,r,ind
                # print irrk.shape, bz.shape
                # bz[r[0],r[1],r[2],:]=irrk.reshape(irrk.size)
                bz[ind, :] = irrk.reshape(irrk.size)
        # duplicate coordinates for the interpolation
        bztmp = bz  # .reshape((mesh[0]*mesh[1]*mesh[2], -1))
        # print bztmp
        ndup = 7
        duplicates = [[-1, 0, 0], [1, 0, 0], [0, -1, 0],
                      [0, 1, 0], [0, 0, -1], [0, 0, 1]]
        bztot = numpy.ndarray((ndup, bztmp.shape[0], bztmp.shape[1]))
        bztot[0, :, :] = bztmp
        ctot = numpy.ndarray((ndup, coords.shape[0], coords.shape[1]))
        ctot[0, :, :] = coords
        for i, sh in enumerate(duplicates):
            bztot[i+1, :, :] = bztmp
            ctot[i+1, :, :] = coords+sh
            # print 'coors',coords,coords+[1.0,0,0]
        bztot = bztot.reshape((ndup*bztmp.shape[0], -1))
        ctot = ctot.reshape((ndup*coords.shape[0], -1))
        import scipy.interpolate.interpnd as interpnd
        self.interpolator = interpnd.LinearNDInterpolator(ctot, bztot)
        # sanity check of the interpolation
        sanity = 0.0
        for kpt in evals:
            diff = numpy.ravel(numpy.ravel(
                kpt)-numpy.ravel(self.interpolator([kpt.kpt])))
            sanity = max(sanity, numpy.dot(diff, diff))
        print('Interpolation bias', sanity)

    def plot(self, path=None, npts=50):
        if path is None:
            # ppath=BZPath(self.lattice,self.special_paths[0]+['G',],self.special_points,npts)
            ppath = BZPath(
                self.lattice, self.special_paths[0], self.special_points, npts)
        else:
            ppath = path
        toto = self.interpolator(ppath.path)
        import matplotlib.pyplot as plt
        # print toto.min(),toto.max()
        for b in toto.transpose():
            plt.plot(ppath.xaxis, b)  # , 'b-')
        plt.axhline(self.fermi_energy, color='k', linestyle='--')
        for p in ppath.xlabel:
            plt.axvline(p, color='k', linestyle='-')
            plt.xticks(ppath.xlabel, ppath.symbols)
        plt.show()



"""
This module defines some classes to perform a calculation using BigDFT
using binding (GIBinding) or using system call (SystemCalculator).

"""

# In our case for the class SystemCalculator which uses system calls:
# * We define posinp (equivalent of Atoms)
# * We have a python dictionary for the parameter
# * We define a calculator (equivalent of BFGS which is an Optimizer
#   (a method to optimize))
# Then we perform the method run.
#
# For the class GIBinding using the Gobject Introspection bindings, two methods
# set and update are added.
#
# The goal is to have a light Calculator almost compatible with ASE
# (Atomic Simulation environment, see https://gitlab.com/ase/ase)
# .. todo::
# In a future we add our method to ASE which is at a higher level
# (workflow of simulations).
# :Example:
# >>> from ase import Atoms
# >>> from ase.optimize import BFGS
# >>>  from ase.calculators.nwchem import NWChem
# >>>  from ase.io import write
# >>>  h2 = Atoms('H2',
# >>>             positions=[[0, 0, 0],
# >>>                        [0, 0, 0.7]])
# >>>  h2.calc = NWChem(xc='PBE')
# >>>  opt = BFGS(h2, trajectory='h2.traj')
# >>>  opt.run(fmax=0.02)
# >>>  BFGS:   0  19:10:49    -31.435229     2.2691
# >>>  BFGS:   1  19:10:50    -31.490773     0.3740
# >>>  BFGS:   2  19:10:50    -31.492791     0.0630
# >>>  BFGS:   3  19:10:51    -31.492848     0.0023
# >>>  write('H2.xyz', h2)
# >>>  h2.get_potential_energy()  # ASE's units are eV and Ang
# >>>  -31.492847800329216
##

import os
from futile.Utils import write as safe_print
import BigDFT.Logfiles as Lf


class GIBinding():
    """
    Calculator for BigDFT from Gobject Introspection bindings.
    """

    def __init__(self):
        # Import bindings about BigDFT (if the bindings are not generated, do
        # not work at all)
        from gi.repository import BigDFT
        self.runObj = -1
        # MPI initialisation
        (ierr, self.iproc, self.nproc, igroup, ngroup) = BigDFT.lib_init(0)
        self.runObj = None

    def update(self, inputfile):
        # If the inputpsiid is not present in the inputfile
        # assumes that the user wants to do a restart
        from futile.Utils import dict_merge
        if "dft" in inputfile and "inputpsiid" in inputfile["dft"]:
            var = inputfile
        else:
            var = inputfile.copy()
            dict_merge(var, {'dft': {'inputpsiid': 1}})
        from gi.repository import BigDFT
        self.runObj.update(BigDFT.Dict(var))

    def run(self):
        self.out = self.runObj.calculate(self.iproc, self.nproc)
        return self.out

    def set(self, inputfile=None):
        from gi.repository import BigDFT
        if inputfile is None:
            var = {}
        else:
            var = inputfile
        # Free memory first
        self.out = None
        self.runObj = None
        self.runObj = BigDFT.Run.new_from_dict(BigDFT.Dict(var))

    def __del__(self):
        if self.runObj == -1:
            return
        # MPI finalisation.
        self.out = None
        self.runObj = None
        from gi.repository import BigDFT
        BigDFT.lib_finalize()


class Runner():
    """Run of something.

    This object is associated with the concept of execution of a action.
    It may be customized to be used inside workflows and datasets.
    The central functionality is the `run` method that can be customized on
    subclasses of `Runner`. In this object there are global and local options
    of a run method. All arguments passed at the instantiation are stored as
    global options. For each call to `run`, these global options may updated by
    the arguments of the run call.

    Args:
        **kwargs: global options of the runner. Deepcopied in the dictionary
          returned by :meth:`global_options`.

    Example:

        >>> torun=Runner(args1='one',args2='two')
        >>> print(torun.global_options())
        {'args1':'one','args2':'two'}
        >>> print(torun.get_global_option('args1'))
        'one'

    """

    def __init__(self, **kwargs):
        import copy
        self._global_options = copy.deepcopy(kwargs)

    def global_options(self):
        """
        Get all global options dict.

        Returns:
            :py:class:`dict`: The dictionary of the global options in its
            current status
        """
        return self._global_options

    def get_global_option(self, key):
        """

        Get one key in global options

        Args:
           key (string): the global option key

        Returns:
            The value of the global options labelled by ``key``

        """
        return self._global_options[key]

    def update_global_options(self, **kwargs):
        """
        Update the global options by providing keyword arguments.

        Args:
           **kwargs: arguments to be updated in the global options
        """
        self._global_options.update(kwargs)

    def pop_global_option(self, key):
        """
        Remove a given global option from the global option dictionary

        Args:
           key (string): the global option key

        Returns:
           The value of the global option
        """
        self._global_options.pop(key)

    def _run_options(self, **kwargs):
        """
        Create a local dictionary for a specific run.
        It combines the present status of global option with the local
        dictionary of the run
        """
        import copy
        # First deepcopy from global_options and update from kwargs (warning: a
        # dictionary is not update)
        self.run_options = copy.deepcopy(self._global_options)
        """ :py:class`dict`: Local options of process_run.
         This dictionary can be accessed during the definition of the
         process_run method.
        It contains all the relevant keys for the definition of the runner.
        """

        self.run_options.update(kwargs)

    def run(self, **kwargs):
        """
        Run method of the class. It performs the following actions:

         * Constructs the local dictionary to be passed as ``**kwargs`` to the
           `process_run` function;
         * Calls the :meth:`pre_processing` method (intended to prepare some
           actions associated to the :meth:`process_run` method);
         * Calls :meth:`process_run` with the dictionary returned by
           :meth:`pre_processing` as  `**kwargs`;
         * Update such dictionary with the results returned by
           :meth:`process_run` and call :meth:`post_processing`;
         * Returns the object passed by the call to :meth:`post_processing`
           class method

        Developers are therefore expected to override :meth:`pre_processing`
        :meth:`process_run` and :meth:`post_processing`,
        when subclassing :class:`Runner`.

        """
        from futile.Utils import dict_merge
        self._run_options(**kwargs)
        run_args = self.pre_processing()
        run_results = self.process_run(**run_args)
        # safe_print('run_args',run_args,'run_results',run_results)
        dict_merge(dest=run_args, src=run_results)
        # safe_print('run_updated, again',run_args)
        return self.post_processing(**run_args)

    def pre_processing(self):
        """
        Pre-treat the keyword arguments and the options, if needed.

        Returns:
           :py:class:`dict`: dictionary of the pre-treated keyword arguments
           that have to be actually considered by process_run.
        """
        return {}

    def process_run(self, **kwargs):
        """
        Main item of the runner, defines the information that have to be
        post_processed by post_processing.

        Args:
          **kwargs (:py:class:`dict`): keyword arguments as returned from the
            :meth:`pre_processing` method.

        Returns:
          :py:class:`dict`:
               dictionary objects to be passed to post_processing, once the
               dictionary returned by :meth:`pre_processing` has been updated
        """
        return kwargs

    def post_processing(self, **kwargs):
        """
        Post-processing, take the arguments as they are provided by the update
        of :meth:`process_run` and :meth:`pre_processing` methods.

        Returns:
           The final object that each call to the :meth:`run` method is
           supposed to provide.
        """
        return None


class SystemCalculator(Runner):
    """Define a BigDFT calculator.

    Main calculator of BigDFT code. It performs :py:meth:`os.system` calls to
    the main ``bigdft`` executable in the ``$BIGDFT_ROOT`` directory. It is
    designed for two purposes:

      * Run the code in a workstation-based environment, for example within
        notebooks or scripts.
      * Run the code from a python script that is submitted to a batch
        scheduler in a potnentially large-scale supercomputer.

    For triggering the execution, this code gets two variables from the
        environment:

        * The value of ``OMP_NUM_THREADS`` to set the number of
          OMP_NUM_THREADS. If this variable is not present in the environment,
          :class:`SystemCalculator` sets it to the value provided by the
          ``omp`` keyword at initialization.

        * The value of ``BIGDFT_MPIRUN`` to define the MPI execution command.
          If absent, the run is executed simply by ``$BIGDFT_ROOT/bigdft``,
          followed by the command given by post-processing.

    Arguments:
         omp (int): number of OpenMP threads.
            It defaults to the $OMP_NUM_THREADS variable in the environment, if
            present, otherwise it fixes the run to 1 thread.
         mpi_run (str): define the MPI command to be used.
            It defaults to the value $BIGDFT_MPIRUN of the environment, if
            present. When using this calculator into a job submission script,
            the value of $BIGDFT_MPIRUN variable may be set appropriately to
            launch the bigdft executable.
         skip (bool): if ``True``, do not run the calculation if the
           corresponding logfile exists.
         verbose (bool): if ``True`` the class prints out informations about
           the operations that are being performed by the calculator
         dry_run (bool): check the input, estimate the memory but do not
           perform the calculation.
         dry_mpi (int): Number of MPI processes for the estimation of the
           memory when ``dry_run`` is ``True`` (not yet implemented)
         taskgroup_size (int): number of MPI processes of each of the
           taskgroup in the case of a runs_file.

    Warning:
       At the initialization, `SystemCalculator` checks if the environment
         variable $BIGDFT_ROOT is defined.
       This would mean (although not guarantee) that the environment has been
         properly set prior to the evaluation of the python command.
       Also, it checks that the executable file ``bigdft`` might be found in
         the ``$BIGDFT_ROOT/bigdft`` path.

    Example:
        >>> inpdict = { 'dft': { 'ixc': 'LDA' }} #a simple input file
        >>> study = SystemCalculator(omp=1)
        >>> logf = study.run(name="test",input=inpdict)
        Executing command:  $BIGDFT_MPIRUN <path_to_$BIGDFT_ROOT>/bigdft test



    Methods:

         run(name='',run_dir='.',outdir='',run_names='',input=None,posinp='posinp.xyz'):

                Run a calculation building the input file from a dictionary.

                Args:
                   name (str): naming scheme of the run i.e. <name>.yaml is
                        the input file and log-<name>.yaml the output one.
                        Data will then be written in the directory
                        `data-<name>.yaml`, unless the "radical" keyword is
                        specified in the input dictionary.
                   run_dir (str): specify the directory where bigdft will be
                   executed (the input and log file will be created in it)
                             It can be a recursive directory path.
                   outdir (str): specify the output directory for all data
                            coming from bigdft
                   run_names (str): File containing the list of the run ids
                            which have to be launched independently
                            (list in yaml format). This option is not
                            compatible with the ``name`` option.
                   input (:py:class:`dict`): give the input parameters (a
                   dictionary or a list of dictionary). If this parameter is
                           absent it is assumed that an inputfile named
                           `name`.yaml exists in the directory indicated by
                           `run_dir`
                   posinp (file or :py:class:`dict`): indicate the posinp file
                           (atomic position file). It can be either a path or a
                           dictionary in the yaml format.

                Returns:
                     Logfile: Instance of the logfile associated to the run.
                     associated to the run which has been just performed.
                     If the run failed for some reasons, the logfile seem not
                     existing or it cannot be parsed it returns `None`.

                Raises:
                     ValueError: if the logfile does not exists or is not
                     accessible, or if the posinp file does not exists

                Todo:
                     Set the return value of run in the case of a run_file. It
                     should be a list of Logfile classes
    """
    import os
    import shutil

    def __init__(self,
                 omp=os.environ.get('OMP_NUM_THREADS', '1'),
                 mpi_run=os.environ.get('BIGDFT_MPIRUN', ''),
                 dry_run=False, skip=False, verbose=True):
        # Use the initialization from the Runner class (so all options inside
        # __global_options)
        Runner.__init__(self, omp=str(omp), mpi_run=mpi_run,
                        dry_run=dry_run, skip=skip, verbose=verbose)
        # Verify if $BIGDFT_ROOT is in the environment
        assert 'BIGDFT_ROOT' in os.environ
        executable = os.path.join(os.environ['BIGDFT_ROOT'], 'bigdft')
        # the bigdft file should be present in the BIGDFT_ROOT directory
        assert os.path.isfile(executable)
        # Build the command setting the number of omp threads
        self.command = (
            self._global_options['mpi_run'] + ' ' + executable).strip()
        safe_print(
            'Initialize a Calculator with OMP_NUM_THREADS=%s and command %s' %
            (self._global_options['omp'], self.command))

    def pre_processing(self):
        # def run(self, name='', outdir='', run_name='', input={},
        #         posinp=None,**kwargs):
        """
        Process local run dictionary to create the input directory and identify
        the command to be passed

        Returns:
            :py:class:`dict`: dictionary containing the command to be passed to
            :meth:`process_run`

        """
        from futile.Utils import make_dict
        from futile import YamlIO as Y
        self._ensure_run_directory()
        # Create the input file (deepcopy because we modify it)
        inp = self.run_options.get('input', {})
        # from here onwards the local input is a dict and not anymore anothe
        # class
        local_input = make_dict(inp)
        # Add into the dictionary a posinp key
        posinp = self.run_options.get('posinp', None)
        if posinp is not None:
            local_input['posinp'] = self._posinp_dictionary_value(posinp)
        # Creating the yaml input file

        if 'input'in self.run_options:
            input_file = self._get_inputfilename()
        elif 'run_names' not in self.run_options:
            input_file = os.path.join(self.run_dir, 'default.yaml')

        if self.run_options['verbose']:
            safe_print('Creating the yaml input file "%s"' % input_file)
        Y.dump(local_input, filename=input_file)
        return {'command': self._get_command()}

    def process_run(self, command):
        """Finally launch the code.

        Routine associated to the running of the ``bigdft`` executable.

        Arguments:
           command (str): the command as it is set by the ``pre_processing``
             method.

        Returns:
           :py:class:`dict`: The dictionary containing `timedbg` and `logname`
             values to be passed to `post_processing` function
        """
        # check if the debug file will be updated (case of erroneous run)
        timedbg = self._get_debugfile_date()
        verbose = self.run_options['verbose']
        # Set the number of omp threads only if the variable is not present
        # in the environment
        if 'OMP_NUM_THREADS' not in os.environ:
            os.environ['OMP_NUM_THREADS'] = self.run_options['omp']
        if verbose:
            if self.run_dir != '.':
                safe_print('Run directory', self.run_dir)
            safe_print('Executing command: ', command)
        # Run the command
        os.system("cd " + self.run_dir + "; " + command)
        return {'timedbg': timedbg, 'logname': self._get_logname()}

    def post_processing(self, timedbg, logname, command):
        """
        Check the existence and the log file and return an instance logfile.

        Returns:
               (Logfile) Instance of the logfile associated to the run.
               associated to the run which has been just performed.
               If the run failed for some reasons, the logfile seem not
               existing or it cannot be parsed it returns `None`.

        Raises:
               ValueError: if the logfile does not exists or is not accessible

        Todo:
           Set the return value of run in the case of a run_file. It should be
           a list of Logfile classes

        """
        # verify that no debug file has been created
        if self._get_debugfile_date() > timedbg:
            verbose = self.run_options['verbose']
            if verbose:
                safe_print(
                    "ERROR: some problem occured during the execution of the"
                    " command, check the 'debug/' directory and the logfile")
                # the debug file is sane, we may print out the error message
                self._dump_debugfile_info()
            try:
                return Lf.Logfile(logname)
            except IOError:
                return None
        if os.path.exists(logname):
            from futile.Utils import file_time
            inputname = self._get_inputfilename()
            if file_time(logname) < file_time(inputname) and not \
               self.run_options['skip']:
                safe_print("ERROR: The logfile (", logname,
                           ") is older than the inputfile (", inputname, ").")
                return None
            else:
                return Lf.Logfile(logname)
        else:
            raise ValueError("The logfile (", logname, ") does not exist.")

    def _get_command(self):
        name = self.run_options.get('name', '')
        dry_run = self.run_options['dry_run']
        run_names = self.run_options.get('run_names', '')
        outdir = self.run_options.get('outdir', '')
        taskgroup_size = self.run_options.get('taskgroup_size', '')
        # Check if it is a dry run
        if dry_run:
            # Use bigdft-tool (do not use BIGDFT_MPIRUN because it is a python
            # script)
            command = os.path.join(
                os.environ['BIGDFT_ROOT'], 'bigdft-tool') + \
                ' -a memory-estimation -l'
            if name > 0:
                command += ' --name=' + name
        else:
            # Adjust the command line with options
            command = self.command
            if name:
                command += ' -n ' + name
            if run_names:
                command += ' -r ' + run_names
            if outdir:
                command += ' -d ' + outdir
            if taskgroup_size:
                command += ' -t ' + taskgroup_size
            if self.run_options['skip']:
                command += ' -s Yes'
        return command

    def _get_logname(self):
        import os
        outdir = self.run_options.get('outdir', '')
        name = self.run_options.get('name', '')
        logname = 'log-' + name + '.yaml' if name else 'log.yaml'
        if outdir:
            logname = os.path.join(outdir, logname)
        logname = os.path.join(self.run_dir, logname)
        return logname

    def _get_inputfilename(self):
        import os
        name = self.run_options.get('name', '')
        input_file = name + '.yaml' if name else 'input.yaml'
        return os.path.join(self.run_dir, input_file)

    def _ensure_run_directory(self):
        from futile.Utils import ensure_dir
        run_dir = self.run_options.get('run_dir', '.')
        # Restrict run_dir to a sub-directory
        if ("/" in run_dir or run_dir == ".."):
            raise ValueError(
                "run_dir '%s' where bigdft is executed must be a sub-directory"
                % run_dir)
        # Create the run_dir if not exist
        if ensure_dir(run_dir) and self.run_options['verbose']:
            safe_print("Create the sub-directory '%s'" % run_dir)

        self.run_dir = run_dir
        """Run directory.
        str: the directory where the inputfile has been copied to.
              Might be useful to associate to each of the calculation of a
              given run a different directory. Note that this is different than
              setting the ``outdir`` or the ``name`` arguments at it refers
              to the directory of the inputfile.
        Note:
            This is not a global property of the calculator, as the same
            calculator instance might be used for various workflows.
        """

    def _posinp_dictionary_value(self, posinp):
        """
        Create the dictionary value associated to posinp field

        Args:
          posinp (str, dict): path of the posinp file. Might be relative or
          absolute. Copied into `run_dir` if not existing. If it is a
          dictionary, it is a representation of the atomic position.

        Returns:
          str,dict: the value of the key ``posinp`` of the input file, if
          posinp is a string, otherwise the posinp dictionary
        """
        import os
        from futile.Utils import ensure_copy, make_dict
        if isinstance(posinp, dict):
            return make_dict(posinp)
        # Check if the file does exist
        if not os.path.isfile(posinp):
            raise ValueError(
                "posinp: The atomic position file '%s' does not exist"
                % posinp)
        posinpdict = posinp
        posinpfile = os.path.basename(posinp)
        # Copy the posinp if not identical
        cp_posinp = os.path.join(self.run_dir, posinpfile)
        copied = ensure_copy(src=posinp, dest=cp_posinp)
        if copied:
            posinpdict = posinpfile
            if self.run_options['verbose']:
                safe_print("Copy the posinp file '%s' into '%s'" %
                           (posinp, self.run_dir))
        return posinpdict

    def _get_debugfile_date(self):
        """
        Get the information about the debug time of the last file in the
        current directory
        """
        from futile.Utils import file_time
        return file_time(os.path.join(self.run_dir, 'debug',
                                      'bigdft-err-0.yaml'))

    def _dump_debugfile_info(self):
        from futile import YamlIO as Y
        debugfile = os.path.join(self.run_dir, 'debug', 'bigdft-err-0.yaml')
        if os.path.isfile(debugfile):
            debugdict = Y.load(debugfile, doc_lists=False)
            safe_print('The error occured is', self._get_error_key(debugdict))
            safe_print('Additional Info: ', debugdict['Additional Info'])

    def _get_error_key(self, debugdict):
        for key in debugdict:
            if 'Calling sequence' in key:
                continue
            if 'Global dictionary' in key:
                continue
            if 'Additional Info' in key:
                continue
            return key


# Test the calculators
if __name__ == '__main__':

    import yaml
    import matplotlib.pyplot as plt

    basicinput = """
#mode: {method: lj}
logfile: No
dft: { ixc: HF, nspin: 2}
posinp:
   positions:
   - {Be : [0.0, 0.0, 0.0]}#, IGSpin: -1}
   - {Be : [0.0, 0.0, 1.0]}#, IGSpin: 1}
#   properties: {format: yaml}
ig_occupation:
   Atom 1: {2s: {up: 1.0, down: 0.9}, 2p: {up: 0.0, down: 0.2} }
   Atom 2: {2s: {up: 0.9, down: 1.0}, 2p: {up: 0.2, down: 0.0} }

psppar.Be: {Pseudopotential XC: 11}
"""

    # Initialize the calculator
    study = GIBinding()
    inp = yaml.load(basicinput)
    study.set(inp)

    # Perform the first calculation
    out = study.run()
    safe_print('starting energy', out.eKS)
    energy = [out.eKS]
    pos = [1.0]

    # Perform a dissociation curve
    for i in range(10):
        sh = float(i + 1) * 0.02
        inp['posinp']['positions'][-1]['Be'][2] += sh
        study.update(inp)
        out = study.run()
        energy.append(out.eKS)
        pos.append(pos[-1] + sh)
        if study.iproc == 0:
            safe_print('iter', i, 'shift', sh, 'energy', out.eKS)
    out = None
    safe_print('End of the calculations')

    # Plot the dissociation curve
    if study.iproc == 0:
        plt.plot(pos, energy)
        plt.show()



from futile.Utils import write as safe_print

def set_inputfile(hgrid, dico, init_input=None, units="reduced"):
    basicinput = """
logfile: Yes
dft:
    ixc: PBE
    ncong: 2
    rmult: [10, 8]
    itermax: 3
    idsx: 0
    gnrm_cv: 1e-8
#Control the diagonalisation scheme
mix:
    iscf: 7
    itrpmax: 200
    rpnrm_cv: 1e-12
    tel: 1e-3
    alphamix: 0.5
    norbsempty: 1000
    alphadiis: 1.0

#perf:
#    accel: OCLGPU
#    ocl_devices: Tesla K40c
#    blas: Yes

"""
    import yaml
    import os
    if not init_input:
        var = yaml.load(basicinput)
    else:
        var = init_input
    # Spin parameters
    var["dft"].update(set_spin(dico["name"], dico["nat"]))
    # K point parameters
    var["kpt"] = set_kpoints(dico["nat"])
    var["dft"]["hgrids"] = (hgrid, hgrid, hgrid)
    # Control the diagonalisation scheme
    if dico["name"] in ("Cr", ):
        var["mix"]["iscf"] = 3
        var["mix"]["alphamix"] = 0.9
    if dico["name"] in ("Ba", "Ca"):
        var["mix"]["norbsempty"] = 8
    var["ig_occupation"] = {dico["name"]: {"empty_shells": ("s", "p", "d")}}
    # Atoms
    pspfile = "psppar."+dico["name"]
    if not os.path.isfile(pspfile):
        safe_print("WARNING: Using default PSP for atom", dico["name"])
    else:
        var[pspfile] = open(pspfile, 'r').read()
    # var["posinp"] = {"positions": [{dico["name"]:
    #                  map(float, dico[i + 1].split())}
    #                  for i in range(dico["nat"])],
    #                  "units": "reduced",
    #                  "cell": (dico["a"], dico["b"], dico["c"])}
    if units == "reduced":
        positions = [{dico["name"]: dico[i + 1]} for i in range(
            dico["nat"])]
    else:
        positions = [{dico["name"]: [dico[i + 1][0]*dico["a"], dico[i + 1][1]*dico["b"],dico[i + 1][2]*dico["c"]]} for i in range(
            dico["nat"])]

    var["posinp"] = {"positions": positions, "units": units,
        "cell": (dico["a"], dico["b"], dico["c"])}
    # We round robin the igspins.
    if "mpol" in var["dft"]:
        mpol = 0
        while mpol < var["dft"]["mpol"]:
            for at in var["posinp"]["positions"]:
                if mpol < var["dft"]["mpol"]:
                    if "IGSpin" in at:
                        at["IGSpin"] += 1
                    else:
                        at["IGSpin"] = 1
                    mpol += 1
    elif "nspin" in var["dft"] and var["dft"]["nspin"] == 2:
        for (i, at) in enumerate(var["posinp"]["positions"]):
            at["IGSpin"] = 1 - 2 * (i % 2)
    return var


def set_spin(name, nat):
    "Define the spin in function of the nature of the atoms"
    dspin = {}
    if name == 'O':
        dspin["nspin"] = 2
    elif name == 'Cr' or name == 'Mn':
        dspin["nspin"] = 2
    elif name == 'Fe' or name == 'Co' or name == 'Ni':
        dspin["nspin"] = 2
        mpol = {"Fe": 2.22, "Co": 1.72, "Ni": 0.60}
        dspin["mpol"] = int(mpol[name] * nat)
        if dspin["mpol"] % 2 == 1:
            dspin["mpol"] += 1
    else:
        dspin["nspin"] = 1
    return dspin


def set_kpoints(nat):
    "Define the k point mesh"
    dkpt = {}
    dkpt["method"] = "mpgrid"
    if nat == 1:
        dkpt["ngkpt"] = [19, 19, 19]
    elif nat == 2:
        dkpt["ngkpt"] = [15, 15, 15]
    elif nat == 3:
        dkpt["ngkpt"] = [14, 14, 14]
    elif nat == 4:
        dkpt["ngkpt"] = [12, 12, 12]
    elif nat == 6:
        dkpt["ngkpt"] = [11, 11, 11]
    elif nat == 8:
        dkpt["ngkpt"] = [10, 10, 10]
    elif nat == 12:
        dkpt["ngkpt"] = [9, 9, 9]
    return dkpt


def set_strain(strain, ngrids, dico):
    return {"radical": dico["name"] + "-%s" % strain,
            "posinp": {"cell": (strain * dico["a"],
                                strain * dico["b"],
                                strain * dico["c"])},
            "dft": {"hgrids": ((strain * dico["a"] + 1e-4) / ngrids[0],
                               (strain * dico["b"] + 1e-4) / ngrids[1],
                               (strain * dico["c"] + 1e-4) / ngrids[2])}}


def set_restart():
    return {"mix": {"tel": 1.e-5}, "dft": {"inputpsiid": 1}}


def transform_to_orthorombic(dico):
    import math
    btype = None
    # Look for possible orthorombic transformation :
    if (dico["alpha"] != 90 and dico["b"] == dico["c"] and
            dico["beta"] == 90 and dico["gamma"] == 90):
        btype = "hexagonal"
        la = "b"
        lb = "c"
        ang = "alpha"
        P = ((1.,  0.,  0.),
             (0.,  0.5, 0.5),
             (0., -0.5, 0.5))
        du = 0.
        dv = 0.
        dw = 1.
    elif (dico["beta"] != 90 and dico["a"] == dico["c"] and
          dico["alpha"] == 90 and dico["gamma"] == 90):
        la = "a"
        lb = "c"
        btype = "hexagonal"
        ang = "beta"
        P = ((0.5, 0., 0.5),
             (0.,  1., 0.),
             (-0.5, 0., 0.5))
        du = 0.
        dv = 0.
        dw = 1.
    elif (dico["gamma"] != 90 and dico["a"] == dico["b"] and
          dico["alpha"] == 90 and dico["beta"] == 90):
        la = "a"
        lb = "b"
        btype = "hexagonal"
        ang = "gamma"
        P = ((0.5, 0.5, 0.),
             (-0.5, 0.5, 0.),
             (0.,  0.,  1.))
        du = 0.
        dv = 1.
        dw = 0.
    elif (dico["gamma"] == 90 and dico["alpha"] == 90 and
          dico["beta"] == 90):
        btype = "orthorombic"
    elif (dico["a"] == dico["b"] and dico["a"] == dico["c"] and
          dico["alpha"] == dico["beta"]and dico["alpha"] == dico["gamma"]):
        btype = "rhombohedral"
    # Transform to orthorombic when possible.
    if btype == "hexagonal":
        a = dico[la]
        b = dico[lb]
        # - hexagonal case.
        alpha = float(dico[ang]) * math.pi / 180.
        dico[la] = math.sqrt((a+b*math.cos(alpha)) **
                             2 + (b*math.sin(alpha))**2)
        dico[lb] = math.sqrt((a-b*math.cos(alpha)) **
                             2 + (b*math.sin(alpha))**2)
        for i in range(dico["nat"]):
            u, v, w = dico[i+1]
            a = P[0][0] * u + P[0][1] * v + P[0][2] * w
            b = P[1][0] * u + P[1][1] * v + P[1][2] * w
            c = P[2][0] * u + P[2][1] * v + P[2][2] * w
            dico[i + 1] = (a, b, c)
            u += du
            v += dv
            w += dw
            a = P[0][0] * u + P[0][1] * v + P[0][2] * w
            b = P[1][0] * u + P[1][1] * v + P[1][2] * w
            c = P[2][0] * u + P[2][1] * v + P[2][2] * w
            dico[dico["nat"] + i + 1] = (a, b, c)
        dico["nat"] *= 2
        dico[ang] = "90"
    elif btype == "rhombohedral" and dico["alpha"] == 60:
        a = dico["a"]
        dico["a"] = a * math.sqrt(2.)
        dico["b"] = a * math.sqrt(2.)
        dico["c"] = a * math.sqrt(2.)
        P = ((0.,  0.5, 0.5),
             (0.5, 0.,  0.5),
             (0.5, 0.5, 0.))
        dd = ((0., 0., 0.),
              (1., 0., 0.),
              (0., 1., 0.),
              (0., 0., 1.))
        nat = dico["nat"]
        dico["nat"] *= len(dd)
        for i in range(nat):
            for (j, (du, dv, dw)) in enumerate(dd):
                u, v, w = dico[i+1]
                u += du
                v += dv
                w += dw
                a = P[0][0] * u + P[0][1] * v + P[0][2] * w
                b = P[1][0] * u + P[1][1] * v + P[1][2] * w
                c = P[2][0] * u + P[2][1] * v + P[2][2] * w
                dico[j * nat + i + 1] = (a, b, c)
                dico["alpha"] = 90.0
                dico["beta"] = 90.0
                dico["gamma"] = 90.0
    elif btype != "orthorombic":
        safe_print("to be treated", dico["name"], dico["alpha"],
                   dico["beta"], dico["gamma"], dico["a"], dico["b"],
                   dico["c"])
